"""Frames methods for pyorc."""

import copy
from typing import Literal, Optional

import cv2
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from ffpiv import window
from matplotlib.animation import FuncAnimation
from tqdm import tqdm

from pyorc import const, cv, helpers, project
from pyorc.velocimetry import ffpiv, openpiv

from .orcbase import ORCBase
from .plot import _frames_plot

__all__ = ["Frames"]


@xr.register_dataarray_accessor("frames")
class Frames(ORCBase):
    """Frames functionalities that can be applied on xr.DataArray."""

    def __init__(self, xarray_obj):
        """Initialize a frames `xr.DataArray`.

        Parameters
        ----------
        xarray_obj: xr.DataArray
            frames data fields (from pyorc.Video.get_frames)

        """
        super(Frames, self).__init__(xarray_obj)

    def get_piv_coords(
        self, window_size: tuple[int, int], search_area_size: tuple[int, int], overlap: tuple[int, int]
    ) -> tuple[dict, dict]:
        """Get Particle Image Velocimetry (PIV) coordinates and mesh grid projections.

        This function calculates the PIV coordinates and the corresponding mesh grid
        projections based on the provided window size, search area size, and overlap.
        The results include both projected coordinates (xp, yp) and the respective
        longitude and latitude values (if available).

        Parameters
        ----------
        window_size : (int, int)
            The size of the window for the PIV analysis.
        search_area_size : (int, int)
            The size of the search area for the PIV analysis.
        overlap : (int, int)
            The overlap ratio between consecutive windows.

        Returns
        -------
        coords : dict
            A dictionary containing the PIV local non-geographical projection coordinates:
                - 'y': The y-axis coordinates.
                - 'x': The x-axis coordinates.
        mesh_coords : dict
            A dictionary containing the mesh grid projections and coordinates:
                - 'xp': The projected x-coordinates.
                - 'yp': The projected y-coordinates.
                - 'xs': The x-coordinates in the image space.
                - 'ys': The y-coordinates in the image space.
                - 'lon': The longitude values (if CRS is provided).
                - 'lat': The latitude values (if CRS is provided).

        """
        dim_size = self._obj[0].shape

        # get the cols and rows coordinates of the expected results
        cols_vector, rows_vector = window.get_rect_coordinates(
            dim_size=dim_size,
            window_size=window_size,
            search_area_size=search_area_size,
            overlap=overlap,
        )
        cols, rows = np.meshgrid(cols_vector, rows_vector)
        # retrieve the x and y-axis belonging to the results
        x, y = helpers.get_axes(cols_vector, rows_vector, self._obj.x.values, self._obj.y.values)
        # convert in projected and latlon coordinates
        xs, ys = helpers.get_xs_ys(
            cols,
            rows,
            self.camera_config.transform,
        )
        if hasattr(self.camera_config, "crs"):
            lons, lats = helpers.get_lons_lats(xs, ys, self.camera_config.crs)
        else:
            lons = None
            lats = None
        # calculate projected coordinates
        z = self.camera_config.h_to_z(self.h_a)
        zs = np.ones(xs.shape) * z
        xp, yp = self.camera_config.project_grid(xs, ys, zs, swap_y_coords=True)
        # package the coordinates
        coords = {"y": y, "x": x}
        mesh_coords = {"xp": xp, "yp": yp, "xs": xs, "ys": ys, "lon": lons, "lat": lats}
        return coords, mesh_coords

    def get_piv(
        self,
        window_size: Optional[tuple[int, int]] = None,
        overlap: Optional[tuple[int, int]] = None,
        engine: str = "numba",
        **kwargs,
    ) -> xr.Dataset:
        """Perform PIV computation on projected frames.

        Only a pipeline graph to computation is set up. Call a result to trigger actual computation. The dataset
        returned contains velocity information for each interrogation window including "v_x" for x-direction velocity
        components, "v_y" for y-direction velocity component, "corr" for the maximum correlation [-] found in the
        interrogation window and s2n [-] for the signal to noise ratio.

        Parameters
        ----------
        window_size : (int, int), optional
            size of interrogation windows in pixels (y, x)
        overlap : (int, int), optional
            amount of overlap between interrogation windows in pixels (y, x)
        engine : str, optional
            select the compute engine, can be "openpiv" (default), "numba", or "numpy". "numba" will give the fastest
            performance but is still experimental. It can boost performance by almost an order of magnitude compared
            to openpiv or numpy. both "numba" and "numpy" use the FF-PIV library as back-end.
        **kwargs : dict
            keyword arguments to pass to the piv engine. For "numba" and "numpy" the argument `chunks` can be provided
            with an integer defining in how many batches of work the total velocimetry problem should be subdivided.

        Returns
        -------
        xr.Dataset
            PIV results in a lazy dask.array form in DataArrays "v_x", "v_y", "corr" and "s2n".

        See Also
        --------
        OpenPIV project: https://github.com/OpenPIV/openpiv-python
        FF-PIV project: https://github.com/localdevices/ffpiv

        """
        camera_config = copy.deepcopy(self.camera_config)
        dt = self._obj["time"].diff(dim="time")
        # Use window_size from camera_config unless provided in the method
        if window_size is not None:
            camera_config.window_size = window_size
        window_size = (
            2 * (camera_config.window_size,)
            if isinstance(camera_config.window_size, int)
            else camera_config.window_size
        )
        # ensure window size is a round number
        window_size = window.round_to_even(window_size)
        search_area_size = window_size
        # set an overlap if not provided in kwargs
        if overlap is None:
            overlap = 2 * (int(round(camera_config.window_size) / 2),)

        # get all required coordinates for the PIV result
        coords, mesh_coords = self.get_piv_coords(window_size, search_area_size, overlap)
        # provide kwargs for OpenPIV analysis
        if engine == "openpiv":
            import warnings

            warnings.warn(
                '"openpiv" is deprecated, please use "numba" or "numpy" as engine',
                DeprecationWarning,
                stacklevel=2,
            )
            kwargs = {
                **kwargs,
                "search_area_size": search_area_size[0],
                "window_size": window_size[0],
                "overlap": overlap[0],
                "res_x": camera_config.resolution,
                "res_y": camera_config.resolution,
            }
            ds = openpiv.get_openpiv(self._obj, coords["y"], coords["x"], dt, **kwargs)
        elif engine in ["numba", "numpy"]:
            kwargs = {
                **kwargs,
                "search_area_size": search_area_size,
                "window_size": window_size,
                "overlap": overlap,
                "res_x": camera_config.resolution,
                "res_y": camera_config.resolution,
            }
            ds = ffpiv.get_ffpiv(self._obj, coords["y"], coords["x"], dt, engine=engine, **kwargs)
        else:
            raise ValueError(f"Selected PIV engine {engine} does not exist.")
        # add all 2D-coordinates
        ds = ds.velocimetry.add_xy_coords(mesh_coords, coords, {**const.PERSPECTIVE_ATTRS, **const.GEOGRAPHICAL_ATTRS})
        # ensure all metadata is transferred
        ds.attrs = self._obj.attrs
        # in case window_size was changed, overrule the camera_config attribute
        ds.attrs.update(camera_config=camera_config.to_json())
        # set encoding
        ds.velocimetry.set_encoding()
        return ds

    def project(
        self,
        method: Literal["cv", "numpy"] = "numpy",
        resolution: Optional[float] = None,
        reducer: Optional[str] = "mean",
    ):
        """Project frames into a projected frames object, with information from the camera_config attr.

        This requires that the CameraConfig contains full gcp information. If a CRS is provided, also "lat" and "lon"
        variables will be added to the output, containing geographical latitude and longitude coordinates.

        Parameters
        ----------
        method : str, optional
            can be `numpy` or `cv` (default `numpy`).
            With `numpy` each individual orthoprojected grid cell is mapped to the image pixels space. For oversampled
            areas, this is also done vice versa. Undersampled areas result in nearest-neighbour interpolations, whilst
            for oversampled, this results in a mean value (if the user uses `mean` as reducer). With `cv` (opencv)
            resampling is performed by first undistorting images, and then by resampling to the
            desired grid. With heavily distorted images and part of the area of interest outside of the field
            of view, the orthoprojection of the corners may end up in the wrong space.
            We recommend using `numpy` and only use cv with simple cases such as nadir-looking videos.
        resolution : float, optional
            resolution to project to. If not provided, this will be taken from the camera config in the metadata
             (Default value = None)
        reducer: str, optional
            Currently only used when `method="numpy"`. Default "mean", resulting in averaging of oversampled grid cells.

        Returns
        -------
        frames: xr.DataArray
             projected frames and x and y in local coordinate system (origin: top-left), lat and lon if a crs was
             provided.

        """
        cc = copy.deepcopy(self.camera_config)
        if resolution is not None:
            cc.resolution = resolution
        # convert bounding box coords into row/column space
        shape = cc.shape
        # prepare y and x axis of targfet
        y = np.flipud(np.linspace(cc.resolution / 2, cc.resolution * (shape[0] - 0.5), shape[0]))
        x = np.linspace(cc.resolution / 2, cc.resolution * (shape[1] - 0.5), shape[1])
        cols, rows = np.meshgrid(np.arange(len(x)), np.arange(len(y)))
        xs, ys = helpers.get_xs_ys(
            cols,
            rows,
            cc.transform,
        )
        if hasattr(cc, "crs"):
            lons, lats = helpers.get_lons_lats(xs, ys, cc.crs)
        else:
            lons = None
            lats = None
        coords = {
            "y": y,
            "x": x,
        }
        ## PROJECTION PREPARATIONS
        # ========================
        z = cc.get_z_a(self.h_a)
        if not (hasattr(project, f"project_{method}")):
            raise ValueError(f"Selected projection method {method} does not exist.")
        proj_method = getattr(project, f"project_{method}")
        da_proj = proj_method(self._obj, cc, x, y, z, reducer)
        # ensure no missing values are persisting
        da_proj = da_proj.fillna(0.0)

        # assign coordinates
        da_proj = da_proj.frames.add_xy_coords(
            {"xs": xs, "ys": ys, "lon": lons, "lat": lats}, coords, const.GEOGRAPHICAL_ATTRS
        )
        if "rgb" in da_proj.dims and len(da_proj.dims) == 4:
            # ensure that "rgb" is the last dimension and dtype is int
            da_proj = da_proj.transpose("time", "y", "x", "rgb")
            da_proj = da_proj.astype("uint8")
        # in case resolution was changed, overrule the camera_config attribute
        da_proj.attrs.update(camera_config=cc.to_json())
        return da_proj

    def normalize(self, samples=15):
        """Remove the temporal mean of sampled frames.

        This is typically used to remove non-moving background from foreground, and helps to increase contrast when
        river bottoms are visible, or when the objective contains partly illuminated and partly shaded parts.

        Parameters
        ----------
        samples : int, optional
            amount of samples to retrieve from frames for estimating standard deviation and mean. Set to a lower
            number to speed up calculation, default: 15 (which is normally sufficient and fast enough).

        Returns
        -------
        xr.DataArray
            normalized frames

        """
        time_interval = round(len(self._obj) / samples)
        assert time_interval != 0, f"Amount of frames is too small to provide {samples} samples"
        # ensure attributes are kept
        xr.set_options(keep_attrs=True)
        mean = self._obj[::time_interval].mean(axis=0).load().astype("float32")
        frames_reduce = self._obj.astype("float32") - mean
        frames_min = frames_reduce.min(axis=-1).min(axis=-1)
        frames_max = frames_reduce.max(axis=-1).max(axis=-1)
        frames_norm = ((frames_reduce - frames_min) / (frames_max - frames_min) * 255).astype("uint8")
        return frames_norm

    def edge_detect(self, wdw_1=1, wdw_2=2):
        """Highlight edges of frame intensities, using a band convolution filter.

        The filter uses two slightly differently convolved images and computes their difference to detect edges.

        Parameters
        ----------
        wdw_1 : int, optional
            window size to use for first gaussian blur filter (Default value = 2)
        wdw_2 : int, optional
            stride to use for second gaussian blur filter (Default value = 4)

        Returns
        -------
        xr.DataArray
            filtered frames (i.e. difference between first and second gaussian convolution)

        """
        # shape = self._obj[0].shape  # single-frame shape does not change
        # da_convert_edge = dask.delayed(cv._convert_edge)
        stride_1 = wdw_1 * 2 + 1
        stride_2 = wdw_2 * 2 + 1
        return xr.apply_ufunc(
            cv._convert_edge,
            self._obj,
            stride_1,
            stride_2,
            input_core_dims=[["y", "x"], [], []],
            output_core_dims=[["y", "x"]],
            vectorize=True,
            #     exclude_dims=set(("y",)),
            #         kwargs={"stride_1": 5, "stride_2": 9},
            dask="parallelized",
            keep_attrs=True,
        )

    def minmax(self, min=-np.Inf, max=np.Inf):
        """Minimum / maximum intensity filter.

        All pixels will be thresholded to a minimum and maximum value.

        Parameters
        ----------
        min : float, optional
            minimum value to bound intensities to. If not provided, no minimum bound is used.
        max : float, optional
            maximum value to bound intensities to. If not provided, no maximum bound is used.

        Returns
        -------
        xr.DataArray
            Treated frames

        """
        return np.maximum(np.minimum(self._obj, max), min)

    def range(self):
        """Return the range of pixel values through time.

        Returned array does not have a time dimension. This filter is typically used to detect
        widely changing pixels, e.g. to distinguish moving water from land.

        Returns
        -------
        xr.DataArray
            Single image (with coordinates) with minimum-maximum range in time [x, y]

        """
        range_da = (self._obj.max(dim="time", keep_attrs=True) - self._obj.min(dim="time", keep_attrs=True)).astype(
            self._obj.dtype
        )  # ensure dtype out is same as dtype in
        return range_da

    def reduce_rolling(self, samples=25):
        """Remove a rolling mean from the frames.

        Very slow method, so in most cases, it is recommended to use Frames.normalize instead.

        Parameters
        ----------
        samples : int, optional
            number of samples per rolling (Default value = 25)

        Returns
        -------
        da : xr.DataArray
            filtered frames

        """
        roll_mean = self._obj.rolling(time=samples).mean()
        assert len(self._obj) >= samples, f"Amount of frames is smaller than requested rolling of {samples} samples"
        # ensure attributes are kept
        xr.set_options(keep_attrs=True)
        # normalize = dask.delayed(cv2.normalize)
        frames_reduce = self._obj - roll_mean
        frames_thres = np.maximum(frames_reduce, 0)
        # # normalize
        frames_norm = (frames_thres * 255 / frames_thres.max(axis=-1).max(axis=-1)).astype("uint8")
        frames_norm = frames_norm.where(roll_mean != 0, 0)
        return frames_norm

    def time_diff(self, thres=2, abs=False):
        """Apply a difference over time.

        Method subtracts frame 1 from frame 2, frame 2 from frame 3, etcetera.
        This method is very efficient to highlight moving objects when the video is very stable. If the video
        is very unstable this filter may lead to very bad results.

        Parameters
        ----------
        thres : float, optional
            obsolute value intensity threshold to set values to zero when intensity is lower than this threshold
            default: 2.
        abs : boolean, optional
            if set to True (default: False) apply absolute value on result

        Returns
        -------
        da : xr.DataArray
            filtered frames

        """
        frames_diff = self._obj.astype(np.float32).diff(dim="time")
        frames_diff = frames_diff.where(np.abs(frames_diff) > thres)
        frames_diff.attrs = self._obj.attrs
        # frames_diff -= frames_diff.min(dim=["x", "y"])
        frames_diff = frames_diff.fillna(0.0)
        if abs:
            return np.abs(frames_diff)
        return frames_diff

    def smooth(self, wdw=1):
        """Smooth each frame with a Gaussian kernel.

        Parameters
        ----------
        wdw : int, optional
            window height or width applied. if set to 1 (default) then the total window is 3x3 (i.e. 2 * 1 + 1). When
            set to 2, the total window is 5x5 (i.e. 2 * 2 + 1). Very effective to apply before ``Frames.time_diff``.
            The value for ``wdw`` shouild be chosen such that the moving features of interest are not removed from
            the view. This can be based on a visual interpretation of a result.

        Returns
        -------
        da : xr.DataArray
            filtered frames

        """
        stride = wdw * 2 + 1
        f = xr.apply_ufunc(
            cv._smooth,
            self._obj,
            stride,
            input_core_dims=[["y", "x"], []],
            output_core_dims=[["y", "x"]],
            output_dtypes=[np.float32],
            vectorize=True,
            dask="parallelized",
            keep_attrs=True,
        )
        return f

    def to_ani(
        self,
        fn,
        figure_kwargs=const.FIGURE_ARGS,
        video_kwargs=const.VIDEO_ARGS,
        anim_kwargs=const.ANIM_ARGS,
        progress_bar=True,
        **kwargs,
    ):
        """Store an animation of the frames in the object.

        Parameters
        ----------
        fn : str
            path to file to which animation is stored
        figure_kwargs : dict, optional
            keyword args passed to ``matplotlib.pyplot.figure`` (Default value = const.FIGURE_ARGS)
        video_kwargs : dict, optional
            keyword arguments passed to ``save`` method of animation, containing parameters such as the frame rate,
            dpi and codec settings to use. (Default value = const.VIDEO_ARGS)
        anim_kwargs : dict, optional
            keyword arguments passed to ``matplotlib.animation.FuncAnimation``
            to control the animation. (Default value = const.ANIM_ARGS)
        progress_bar : bool, optional
            print a progress bar while storing video (default: True)
        **kwargs : dict
            keyword arguments to pass to ``matplotlib.pyplot.imshow``

        """

        def init():
            # set imshow data to values in the first frame
            im.set_data(self._obj[0])
            return ax  # line,

        def animate(i):
            # set imshow data to values in the current frame
            im.set_data(self._obj[i])
            return ax

        # setup a standard 16/9 borderless window with black background
        f = plt.figure(**figure_kwargs)
        f.set_size_inches(16, 9, True)
        f.patch.set_facecolor("k")
        f.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=None, hspace=None)
        ax = plt.subplot(111)

        im = ax.imshow(self._obj[0], **kwargs)
        if progress_bar:
            frames = tqdm(range(len(self._obj)), position=0, leave=True)
        else:
            frames = range(len(self._obj))
        anim = FuncAnimation(f, animate, init_func=init, frames=frames, **anim_kwargs)
        anim.save(fn, **video_kwargs)

    def to_video(self, fn, video_format=None, fps=None, progress=True):
        """Write frames to a video file without any layout.

        Frames from the input object are written into a video file. The format and frame
        rate can be customized as per user preference or derived automatically from the
        input object.

        Parameters
        ----------
        fn : str
            Path to the output video file.
        video_format : cv2.VideoWriter_fourcc, optional
            The desired video file format codec. If not provided, defaults to
            `cv2.VideoWriter_fourcc(*"mp4v")`.
        fps : float, optional
            Frames per second for the output video. If not specified, it is estimated
            from the time differences in the input frames.
        progress : bool, optional
            Display a progress bar while writing the video frames. (default: True)

        """
        if video_format is None:
            # set to a default
            video_format = cv2.VideoWriter_fourcc(*"mp4v")
        if fps is None:
            # estimate it from the time differences
            fps = 1 / (self._obj["time"].diff(dim="time").values.mean())
        h = self._obj.shape[1]
        w = self._obj.shape[2]
        out = cv2.VideoWriter(fn, video_format, fps, (w, h))
        with tqdm(total=len(self._obj), position=0, leave=True, disable=not (progress)) as pbar:
            pbar.set_description("Writing frames")
            first_frame = True
            for n_start in range(0, len(self._obj), self._obj.chunksize):
                frames_chunk = self._obj.isel(time=slice(n_start, n_start + self._obj.chunksize))
                frames_chunk.load()  # load in memory only once
                for f in frames_chunk:
                    if len(f.shape) == 3:
                        img = cv2.cvtColor(np.uint8(f.values), cv2.COLOR_RGB2BGR)
                    else:
                        img = f.values
                        if first_frame:
                            first_frame = False
                            # make a scale between 0 and 255, only with first frame
                            img_min = img.min(axis=0).min(axis=0)
                            img_max = img.max(axis=0).max(axis=0)
                        img = np.uint8(255 * ((img - img_min) / (img_max - img_min)))
                        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)

                    out.write(img)
                    pbar.update(1)
        out.release()

    plot = _frames_plot



"""Constants used in pyorc."""

import operator
from typing import Literal

# attributes for PIV variables

v_x = "v_x"
v_y = "v_y"
s2n = "s2n"
corr = "corr"

DIST_COEFFS = [[0.0], [0.0], [0.0], [0.0]]

PIV_ATTRS = {
    v_x: {
        "standard_name": "sea_water_x_velocity",
        "long_name": "Flow element center velocity vector, x-component",
        "units": "m s-1",
        "coordinates": "lon lat",
    },
    v_y: {
        "standard_name": "sea_water_x_velocity",
        "long_name": "Flow element center velocity vector, x-component",
        "units": "m s-1",
        "coordinates": "lon lat",
    },
    s2n: {
        "standard_name": "ratio",
        "long_name": "signal to noise ratio",
        "units": "",
        "coordinates": "lon lat",
    },
    corr: {
        "standard_name": "correlation_coefficient",
        "long_name": "correlation coefficient between frames",
        "units": "",
        "coordinates": "lon lat",
    },
}

# attributes for geographical coordinate variables
GEOGRAPHICAL_ATTRS = {
    "xs": {
        "axis": "X",
        "long_name": "x-coordinate in Cartesian system",
        "units": "m",
    },
    "ys": {
        "axis": "Y",
        "long_name": "y-coordinate in Cartesian system",
        "units": "m",
    },
    "lon": {
        "long_name": "longitude",
        "units": "degrees_east",
    },
    "lat": {
        "long_name": "latitude",
        "units": "degrees_north",
    },
}

# attributes for camera perspective coordinate values
PERSPECTIVE_ATTRS = {
    "xp": {"axis": "X", "long_name": "column in camera perspective", "units": "-"},
    "yp": {"axis": "Y", "long_name": "row in camera perspective", "units": "-"},
}

# typical arguments used for making an animation writer.
VIDEO_ARGS = {
    "fps": 10,
    "extra_args": ["-vcodec", "libx264"],
    "dpi": 120,
}
ALLOWED_COLOR_METHODS_WATER_LEVEL = Literal["grayscale", "hue", "sat", "val"]

ANIM_ARGS = {"interval": 20, "blit": False}

ENCODING_PARAMS = {"zlib": True, "dtype": "int16", "scale_factor": 0.01, "_FillValue": -9999}

ENCODE_VARS = [v_x, v_y, corr, s2n]
ENCODING = {k: ENCODING_PARAMS for k in ENCODE_VARS}

FIGURE_ARGS = {
    "figsize": (16, 9),
    "frameon": False,
}


CLASSIFY_MOVING_CAM = [
    {"method": "kmeans", "op": operator.ge},
    {"method": "dist", "q_threshold": 0.5, "op": operator.ge},
]

CLASSIFY_STANDING_CAM = [
    {"method": "kmeans", "op": operator.le},
    {"method": "dist", "q_threshold": 0.8, "op": operator.le},
]

CLASSIFY_CAM = {"fixed": CLASSIFY_STANDING_CAM, "moving": CLASSIFY_MOVING_CAM}

WATER_LEVEL_MAX_DIFF = 20.0



"""Module for camera configuration service."""

import os.path

import matplotlib.pyplot as plt

from pyorc import CameraConfig, Video


def camera_config(
    video_file, cam_config_file, lens_position=None, corners=None, frame_sample=0.0, rotation=None, **kwargs
):
    """Create recipe for preparing a configuration file.

    Parameters
    ----------
    video_file : str,
        Path to file with sample video containing objective of interest
    cam_config_file : str,
        Path to output file containing json camera config
    lens_position : array-like
        x, y, z position of lens in real-world coordinates
    corners : list of lists
        col, row corner points for definition of bounding box
    frame_sample : int
        frame number to use for sampling the control points and bounding box
    rotation : Literal[90, 180, 270]
        rotation in degrees to apply on frame

    **kwargs: dict,
        Keyword arguments to pass to pyorc.CameraConfig (height and width are added on-the-fly from `video_file`)

    Returns
    -------
    None

    """
    # prepare file name for geographical and objective overview
    fn_geo = f"{os.path.splitext(cam_config_file)[0]}_geo.jpg"
    fn_cam = f"{os.path.splitext(cam_config_file)[0]}_cam.jpg"
    # open video for dimensions
    video = Video(video_file, start_frame=frame_sample, end_frame=frame_sample + 1, rotation=rotation)
    # extract first frame
    img = video.get_frame(frame_sample)
    img_rgb = video.get_frame(frame_sample, method="rgb")
    kwargs["height"], kwargs["width"] = img.shape
    # prepare camera config
    cam_config = CameraConfig(rotation=rotation, **kwargs)
    # write to output file
    if lens_position is not None:
        # set lens position assuming the same crs as the gcps
        cam_config.set_lens_position(*lens_position, crs=kwargs["gcps"]["crs"])
    if corners is not None:
        cam_config.set_bbox_from_corners(corners)
    cam_config.to_file(cam_config_file)
    if kwargs["crs"] is not None:
        ax = cam_config.plot(tiles="GoogleTiles", tiles_kwargs={"style": "satellite"})
    else:
        ax = cam_config.plot()
        ax.axis("equal")
    ax.figure.savefig(fn_geo)
    plt.close("all")
    f = plt.figure()
    ax = plt.axes()
    ax.imshow(img_rgb)
    cam_config.plot(ax=ax, mode="camera")
    f.savefig(fn_cam)  # , bbox_inches="tight")



"""Retrieval of sample dataset."""

import os
import zipfile


def get_hommerich_dataset():
    """Retrieve and cache sample dataset of Sheaf river."""
    try:
        import pooch
    except ImportError:
        raise ImportError("This function needs pooch. Install pyorc with pip install pyopenrivercam[extra]")

    # Define the DOI link
    filename = "20241010_081717.mp4"
    base_url = "doi:10.5281/zenodo.15002591"
    url = base_url + "/" + filename
    print(f"Retrieving or providing cached version of dataset from {url}")
    # Create a Pooch registry to manage downloads
    registry = pooch.create(
        # Define the cache directory
        path=pooch.os_cache("pyorc"),
        # Define the base URL for fetching data
        base_url=base_url,
        # Define the registry with the file we're expecting to download
        registry={filename: None},
    )
    # Fetch the dataset
    file_path = registry.fetch(filename, progressbar=True)
    print(f"Hommerich video is available in {file_path}")
    return file_path


def get_hommerich_pyorc_zip():
    """Retrieve and cache sample dataset of Sheaf river."""
    try:
        import pooch
    except ImportError:
        raise ImportError("This function needs pooch. Install pyorc with pip install pyopenrivercam[extra]")

    # Define the DOI link
    filename = "hommerich_20241010_081717_pyorc_data.zip.zip"
    base_url = "doi:10.5281/zenodo.15002591"
    url = base_url + "/" + filename
    print(f"Retrieving or providing cached version of dataset from {url}")
    # Create a Pooch registry to manage downloads
    registry = pooch.create(
        # Define the cache directory
        path=pooch.os_cache("pyorc"),
        # Define the base URL for fetching data
        base_url=base_url,
        # Define the registry with the file we're expecting to download
        registry={filename: None},
    )
    # Fetch the dataset
    file_path = registry.fetch(filename, progressbar=True)
    print(f"Hommerich video is available in {file_path}")
    return file_path


def get_hommerich_pyorc_files():
    """Unzip hommerich pyorc files and return file list."""
    zip_file = get_hommerich_pyorc_zip()
    trg_dir = os.path.split(zip_file)[0]
    if not os.path.exists(os.path.join(trg_dir, "hommerich.yml")):
        # apparently the pyorc files are not yet preset. Unzip the record.
        print("Unzipping sample data...")
        with zipfile.ZipFile(zip_file, "r") as f:
            f.extractall(trg_dir)
    return trg_dir



"""pyorc plot helpers functions."""

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from shapely import geometry


def _plot_3d_pol(polygon, ax=None, **kwargs):
    """Plot single polygon on matplotlib 3d ax."""
    x, y, z = zip(*polygon.exterior.coords)
    verts = [list(zip(x, y, z))]

    if ax is None:
        ax = plt.axes(projection="3d")
    poly = Poly3DCollection(verts, **kwargs)
    # p = ax.plot_trisurf(x, y, z, **kwargs)
    p = ax.add_collection3d(poly)
    return p


def plot_3d_polygon(polygon, ax=None, **kwargs):
    """Plot a shapely.geometry.Polygon or MultiPolygon on matplotlib 3d ax."""
    if isinstance(polygon, geometry.MultiPolygon):
        for pol in polygon.geoms:
            p = _plot_3d_pol(pol, ax=ax, **kwargs)
    else:
        p = _plot_3d_pol(polygon, ax=ax, **kwargs)
    return p


def plot_polygon(polygon, ax=None, **kwargs):
    """Plot a shapely.geometry.Polygon or MultiPolygon on matplotlib ax."""
    if ax is None:
        ax = plt.axes()
    if isinstance(polygon, geometry.MultiPolygon):
        for pol in polygon.geoms:
            patch = plt.Polygon(pol.exterior.coords, **kwargs)
            p = ax.add_patch(patch)
    else:
        patch = plt.Polygon(polygon.exterior.coords, **kwargs)
        p = ax.add_patch(patch)
    return p


def plot_3d_line(line, ax=None, **kwargs):
    """Plot a shapely.geometry.LineString on matplotlib ax."""
    if ax is None:
        ax = plt.axes(projection="3d")
    x, y, z = zip(*line.coords)
    p = ax.plot(x, y, z, **kwargs)
    return p


def plot_line(line, ax=None, **kwargs):
    """Plot a shapely.geometry.LineString on matplotlib ax."""
    if ax is None:
        ax = plt.axes()
    x, y = zip(*line.coords)
    p = ax.plot(x, y, **kwargs)
    return p



"""Velocimetry functionalities that can be applied on `xarray.Dataset`."""

import warnings

import numpy as np
import rasterio
import xarray as xr
from pyproj import CRS
from scipy.interpolate import interp1d
from xarray.core import utils

from pyorc import const, helpers
from pyorc.api.mask import _Velocimetry_MaskMethods
from pyorc.api.orcbase import ORCBase
from pyorc.api.plot import _Velocimetry_PlotMethods


@xr.register_dataset_accessor("velocimetry")
class Velocimetry(ORCBase):
    """Velocimetry functionalities that can be applied on `xarray.Dataset`."""

    def __init__(self, xarray_obj):
        """Initialize a velocimetry `xarray.Dataset`.

        Parameters
        ----------
        xarray_obj: xr.Dataset
            velocimetry data fields (from ``pyorc.Frames.get_piv``)

        """
        super(Velocimetry, self).__init__(xarray_obj)

    @property
    def is_velocimetry(self):
        """Check if the data contained in the object seems to be velocimetry data.

        Done by checking naming of dims and available variables.

        Returns
        -------
        is_velocimetry : bool
            If True, the dataset likely contains velocimetry data

        """
        # check for dims, difference between available and allowed should be zero in length
        unknown_dims = set(self._obj.dims).difference(set(["time", "y", "x"]))
        if len(unknown_dims) != 0:
            print(f"Unknown dimension(s) found: {unknown_dims}")
            return False
        missed_dims = set(["y", "x"]).difference(set(self._obj.dims))
        if len(missed_dims) != 0:
            print(f"Dimensions missing: {missed_dims}")
            return False
        # check for
        missed_vars = set(const.ENCODE_VARS).difference(set(self._obj.data_vars))
        if len(missed_vars) != 0:
            print(f"Variables missing: {missed_vars}")
            return False
        # check for available metadata
        if not (hasattr(self._obj, "camera_config")):
            print("camera_config metadata is missing")
            return False
        return True

    mask = utils.UncachedAccessor(_Velocimetry_MaskMethods)

    def get_transect(
        self,
        x,
        y,
        z=None,
        s=None,
        crs=None,
        v_eff=True,
        xs="xs",
        ys="ys",
        distance=None,
        wdw=1,
        wdw_x_min=None,
        wdw_x_max=None,
        wdw_y_min=None,
        wdw_y_max=None,
        rolling=None,
        tolerance=0.5,
        quantiles=None,
    ):
        """Interpolate all variables to supplied x and y coordinates of a cross-section.

        This function assumes that the grid can be rotated and that xs and ys are supplied following the projected
        coordinates supplied in "xs" and "ys" coordinate variables in ds. x-coordinates and y-coordinates that fall
        outside the domain of ds, are still stored in the result for further interpolation or extrapolation.
        Original coordinate values supplied are stored in coordinates "x", "y" and (if supplied) "z".
        Time series are transformed to set quantiles.

        Parameters
        ----------
        x : tuple or list-like
            x-coordinates on which interpolation should be done
        y : tuple or list-like
            y-coordinates on which interpolation should be done
        z : tuple or list-like
            z-coordinates on which interpolation should be done, defaults: None
        s : tuple or list-like
            distance from bank coordinates on which interpolation should be done, defaults: None
            if set, these distances will be precisely respected, and not interpolated. ``distance`` will be ignored.
        crs : int, dict or str, optional
            coordinate reference system (e.g. EPSG code) in which x, y and z are measured (default: None),
            None assumes crs is the same as crs of xr.Dataset.
        v_eff : boolean, optional
            if True (default), effective velocity (perpendicular to cross-section) is also computed
        xs : str, optional
            name of variable that stores the x coordinates in the projection in which "x" is supplied (default: "xs")
        ys : str, optional
            name of variable that stores the y coordinates in the projection in which "y" is supplied (default: "ys")
        distance : float, optional
            sampling distance over the cross-section in [m]. the bathymetry points will be interpolated to match this
            distance. If not set, the distance will be estimated from the velocimetry grid resolution. (default: None)
        wdw : int, optional
            window size to use for sampling the velocity. zero means, only cell itself, 1 means 3x3 window.
            (default: 1) wdw is used to fill wdw_x_min and wdwd_y_min with its negative (-wdw) value, and wdw_y_min and
            wdw_y_max with its positive value, to create a sampling window.
        wdw_x_min : int, optional
            window size in negative x-direction of grid (must be negative), overrules wdw in negative x-direction if set
        wdw_x_max : int, optional
            window size in positive x-direction of grid, overrules wdw in positive x-direction if set
        wdw_y_min : int, optional
            window size in negative y-direction of grid (must be negative), overrules wdw in negative y-direction if set
        wdw_y_max : int, optional
            window size in positive y-direction of grid, overrules wdw in positive x-direction if set.
        tolerance : float (0-1), optional
            tolerance on the required amount of sampled data in the window defined by wdw and/or wdw_x_min, wdw_x_max,
            wdw_y_min and wdw_y_max (if set). At least this fraction of cells each time step must have a data value
            to return a value. Otherwise the location is given a nan as value.
        rolling : int, optional
            if set other than None (default), a rolling mean over time is applied, before deriving quantile estimates.
        quantiles : list of floats (0-1), optional
            list of quantiles to return (default: [0.05, 0.25, 0.5, 0.75, 0.95]).

        Returns
        -------
        ds_points: xr.Dataset
            interpolated data at the supplied x and y coordinates over quantiles

        """
        if quantiles is None:
            quantiles = [0.05, 0.25, 0.5, 0.75, 0.95]
        transform = helpers.affine_from_grid(self._obj[xs].values, self._obj[ys].values)
        if crs is not None:
            # transform coordinates of cross-section
            x, y = zip(
                *helpers.xyz_transform(list(zip(*(x, y))), crs_from=crs, crs_to=CRS.from_wkt(self.camera_config.crs)),
            )
            x, y = list(x), list(y)
        if s is None:
            if distance is None:
                # interpret suitable sampling distance from grid resolution
                distance = np.abs(np.diff(self._obj.x)[0])
                # interpolate to a suitable set of points
            x, y, z, s = helpers.xy_equidistant(x, y, distance=distance, z=z)

        # make a cols and rows temporary variable
        coli, rowi = np.meshgrid(np.arange(len(self._obj["x"])), np.arange(len(self._obj["y"])))
        self._obj["cols"], self._obj["rows"] = (["y", "x"], coli), (["y", "x"], rowi)
        # compute rows and cols locations of coordinates (x, y)
        rows, cols = rasterio.transform.rowcol(
            transform, list(x), list(y), op=float
        )  # ensure we get rows and columns in fractions instead of whole numbers
        rows, cols = np.array(rows), np.array(cols)

        # compute transect coordinates in the local grid coordinate system (can be outside the grid)
        f_x = interp1d(np.arange(0, len(self._obj["x"])), self._obj["x"], fill_value="extrapolate")
        f_y = interp1d(np.arange(0, len(self._obj["y"])), self._obj["y"], fill_value="extrapolate")
        _x = f_x(cols)
        _y = f_y(rows)

        # covert local coordinates to DataArray
        _x = xr.DataArray(list(_x), dims="points")
        _y = xr.DataArray(list(_y), dims="points")

        # interpolate velocities over points
        if wdw == 0:
            ds_points = self._obj.interp(x=_x, y=_y, method="nearest")
        else:
            # collect points within a stride, collate and analyze for outliers
            ds_wdw = helpers.stack_window(
                self._obj, wdw=wdw, wdw_x_min=wdw_x_min, wdw_x_max=wdw_x_max, wdw_y_min=wdw_y_min, wdw_y_max=wdw_y_max
            )
            # use the median (not mean) to prevent a large influence of serious outliers
            missing_tolerance = ds_wdw.mean(dim="time").count(dim="stride") > tolerance * len(ds_wdw.stride)
            # missing_tolerance = ds_wdw.count(dim="stride") > tolerance*len(ds_wdw.stride)
            ds_effective = ds_wdw.median(dim="stride", keep_attrs=True)
            # remove velocities that are too few in samples
            ds_effective = ds_effective.where(missing_tolerance)
            # scipy does not tolerate np.float32 since scipy=1.10.0, so first convert to np.float64
            for var in ds_effective:
                ds_effective[var] = ds_effective[var].astype(np.float64)
            for coord in ds_effective.coords:
                ds_effective[coord] = ds_effective[coord].astype(np.float64)

            ds_points = ds_effective.interp(x=_x, y=_y)
        if np.isnan(ds_points["v_x"].mean(dim="time")).all():
            warnings.warn(
                "No valid velocimetry points found over bathymetry. Check if the bathymetry is within the camera "
                "objective or anything is visible in objective. ",
                stacklevel=2,
            )
        # add the xcoords and ycoords (and zcoords if available) originally assigned so that even points outside the
        # grid covered by ds can be found back from this dataset
        ds_points = ds_points.assign_coords(xcoords=("points", list(x)))
        ds_points = ds_points.assign_coords(ycoords=("points", list(y)))
        ds_points = ds_points.assign_coords(scoords=("points", list(s)))
        if z is not None:
            ds_points = ds_points.assign_coords(zcoords=("points", list(z)))
        # add mean angles to dataset
        alpha = helpers.xy_angle(ds_points["x"], ds_points["y"])
        flow_dir = alpha - 0.5 * np.pi
        ds_points["v_dir"] = (("points"), flow_dir)
        ds_points["v_dir"].attrs = {
            "standard_name": "river_flow_angle",
            "long_name": "Angle of river flow in radians from North",
            "units": "rad",
        }
        # convert to a Transect object
        ds_points = xr.Dataset(ds_points, attrs=ds_points.attrs)
        if rolling is not None:
            ds_points = ds_points.rolling(time=rolling, min_periods=1).mean()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            ds_points = ds_points.quantile(quantiles, dim="time", keep_attrs=True)
        if v_eff:
            # add the effective velocity, perpendicular to cross-section direction
            ds_points.transect.vector_to_scalar()
        return ds_points

    # add .plot as group of methods, UncachedAccessor ensures that Velocimetry is passed as object
    plot = utils.UncachedAccessor(_Velocimetry_PlotMethods)

    def set_encoding(self, enc_pars=const.ENCODING_PARAMS):
        """Set encoding parameters for all typical variables in a velocimetry dataset.

        This reduces the required storage for this dataset significantly, when stored to disk in e.g. a netcdf file
        using ``xarray.Dataset.to_netcdf``.

        Parameters
        ----------
        enc_pars : dict of dicts, optional
            per variable, a dict containing encoding parameters. When called without input, a standard set of encoding
            parameters is used that compresses well. (Default value = const.ENCODING_PARAMS)

        """
        for k in const.ENCODE_VARS:
            self._obj[k].encoding = enc_pars



"""pyorc.cli Interactive plotting elements."""

try:
    import cartopy.crs as ccrs
    import cartopy.io.img_tiles as cimgt

    use_cartopy = True
except Exception:
    use_cartopy = False

import logging

import click
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patheffects
from matplotlib.backend_bases import MouseButton
from matplotlib.patches import Polygon
from matplotlib.widgets import Button
from mpl_toolkits.axes_grid1 import Divider, Size

from pyorc import helpers
from pyorc.cli import cli_utils

path_effects = [
    patheffects.Stroke(linewidth=2, foreground="w"),
    patheffects.Normal(),
]

corner_labels = ["upstream-left", "downstream-left", "downstream-right", "upstream-right"]


class BaseSelect:
    """Base interactive plot."""

    def __init__(self, img, dst=None, crs=None, buffer=0.0002, zoom_level=19, logger=logging):
        """Set up base interactive plot for camera configs."""
        self.logger = logger
        self.height, self.width = img.shape[0:2]
        if use_cartopy:
            self.crs = crs
        else:
            self.crs = None
        fig = plt.figure(figsize=(16, 9), frameon=False, facecolor="black")
        fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=None, hspace=None)
        ax_geo = None
        if dst is not None:
            xmin = np.array(dst)[:, 0].min()
            xmax = np.array(dst)[:, 0].max()
            ymin = np.array(dst)[:, 1].min()
            ymax = np.array(dst)[:, 1].max()
            extent = [xmin - buffer, xmax + buffer, ymin - buffer, ymax + buffer]
            if self.crs is not None:
                tiler = cimgt.GoogleTiles(style="satellite")
                ax_geo = fig.add_axes([0.0, 0.0, 1, 1], projection=tiler.crs)
                ax_geo.set_extent(extent, crs=ccrs.PlateCarree())
                ax_geo.add_image(tiler, zoom_level, zorder=1)
            else:
                ax_geo = fig.add_axes([0.0, 0.0, 1, 1])
                ax_geo.set_aspect("equal")
            plt.tick_params(left=False, right=False, labelleft=False, labelbottom=False, bottom=False)
            ax_geo.set_visible(False)
        ax = fig.add_axes([0.2, 0.1, 0.7, 0.8])
        ax.set_facecolor("k")
        ax.set_position([0, 0, 1, 1])
        plt.tick_params(left=False, right=False, labelleft=False, labelbottom=False, bottom=False)

        # fig, ax = plt.subplots()
        ax.imshow(img)
        ax.set_title("Left: add point, right: remove point, close: store in .src")
        kwargs = dict(color="w", markeredgecolor="k", markersize=10, zorder=3, label="Control points")
        kwargs_text = dict(
            xytext=(6, 6),
            textcoords="offset points",
            zorder=4,
            path_effects=[
                patheffects.Stroke(linewidth=3, foreground="w"),
                patheffects.Normal(),
            ],
        )
        if dst is not None:
            if self.crs is not None:
                kwargs["transform"] = ccrs.PlateCarree()
                transform = ccrs.PlateCarree()._as_mpl_transform(ax_geo)
                kwargs_text["xycoords"] = transform
            self.p_geo = ax_geo.plot(*list(zip(*dst))[0:2], "o", **kwargs)
            for n, _pt in enumerate(dst):
                _ = ax_geo.annotate(n + 1, xy=_pt[0:2], **kwargs_text)
        self.fig = fig
        self.ax_geo = ax_geo
        self.ax = ax  # add axes
        self.pts_t = []
        self.add_buttons()
        self.press = False
        self.move = False
        self.press_event = self.ax.figure.canvas.mpl_connect("button_press_event", self.on_press)
        self.release_event = self.ax.figure.canvas.mpl_connect("button_release_event", self.on_release)
        self.close_event = self.ax.figure.canvas.mpl_connect("close_event", self.on_close)
        self.move_event = self.ax.figure.canvas.mpl_connect("motion_notify_event", self.on_move)
        self.required_clicks = None  # required clicks should be defined by the respective child class
        self.src = []
        self.dst = dst

    def add_buttons(self):
        """Add interactive buttons for switching axes."""
        h = [Size.Fixed(0.5), Size.Fixed(0.6)]
        v = [Size.Fixed(0.95), Size.Fixed(0.3)]
        v2 = [Size.Fixed(1.45), Size.Fixed(0.3)]
        v3 = [Size.Fixed(1.95), Size.Fixed(0.3)]

        divider3 = Divider(self.fig, (0, 0, 1, 1), h, v, aspect=False)
        divider2 = Divider(self.fig, (0, 0, 1, 1), h, v2, aspect=False)
        divider1 = Divider(self.fig, (0, 0, 1, 1), h, v3, aspect=False)
        if self.ax_geo is not None:
            self.ax_button1 = self.fig.add_axes(divider1.get_position(), axes_locator=divider1.new_locator(nx=1, ny=1))
            self.button1 = Button(self.ax_button1, "Camera")
            self.button1.on_clicked(self.switch_to_ax)
            self.ax_button2 = self.fig.add_axes(divider2.get_position(), axes_locator=divider2.new_locator(nx=1, ny=1))
            self.button2 = Button(self.ax_button2, "Map")
            self.button2.on_clicked(self.switch_to_ax_geo)
        self.ax_button3 = self.fig.add_axes(divider3.get_position(), axes_locator=divider3.new_locator(nx=1, ny=1))
        self.button3 = Button(self.ax_button3, "Done")
        self.button3.on_clicked(self.close_window)
        self.button3.set_active(False)
        self.button3.label.set_color("gray")

    def close_window(self, event):
        """Define close event of entire window."""
        # close window
        plt.close(self.fig)
        # exectute close event
        self.on_close(event)

    def on_close(self, event):
        """Define close event for axes."""
        self.ax.figure.canvas.mpl_disconnect(self.press_event)
        self.ax.figure.canvas.mpl_disconnect(self.release_event)
        self.ax.figure.canvas.mpl_disconnect(self.close_event)
        # check if the amount of src points and dst points is equal. If not return error
        check_length = len(self.src) == self.required_clicks
        if not (check_length):
            raise click.UsageError(
                f"Aborting because you have not supplied all {self.required_clicks} ground control points. "
                f"Only {len(self.src)} points supplied"
            )

    def on_press(self, event):
        """Detect press during click event."""
        self.press = True
        self.move = False

    def on_move(self, event):
        """Detect movement during event."""
        if self.press:
            self.move = True

    def on_click(self, event):
        """Define general click event."""
        # only if ax is visible and click was within the window
        if self.ax.get_visible() and event.inaxes == self.ax:
            if event.button is MouseButton.RIGHT:
                self.on_right_click(event)
            elif event.button is MouseButton.LEFT:
                if len(self.src) < self.required_clicks:
                    self.on_left_click(event)
                # check if enough points are collected to enable the Done button
            if len(self.src) == self.required_clicks:
                self.button3.set_active(True)
                self.button3.label.set_color("k")
            else:
                self.button3.set_active(False)
                self.button3.label.set_color("grey")

        self.ax.figure.canvas.draw()

    def on_left_click(self, event):
        """Define left-click event."""
        if event.xdata is not None:
            self.logger.debug(f"Storing coordinate x: {event.xdata} y: {event.ydata} to src")
            self.src.append([int(np.round(event.xdata)), int(np.round(event.ydata))])
            self.p.set_data(*list(zip(*self.src)))
            pt = self.ax.annotate(
                len(self.src),
                xytext=(6, 6),
                xy=(event.xdata, event.ydata),
                textcoords="offset points",
                zorder=4,
                path_effects=[
                    patheffects.Stroke(linewidth=3, foreground="w"),
                    patheffects.Normal(),
                ],
            )
            self.pts_t.append(pt)

    def on_release(self, event):
        """Define release click event."""
        if self.press and not self.move:
            self.on_click(event)
        self.press = False
        self.move = False

    def on_right_click(self, event):
        """Define right-click event."""
        # remove the last added point
        if len(self.pts_t) > 0:
            self.logger.debug("Removing last generated point")
            self.pts_t[-1].remove()
            del self.pts_t[-1]
        if len(self.src) > 0:
            del self.src[-1]
            if len(self.src) > 0:
                self.p.set_data(*list(zip(*self.src)))
            else:
                self.p.set_data([], [])

    def switch_to_ax(self, event):
        """Switch to camera axes."""
        self.fig.canvas.manager.toolbar.home()
        plt.sca(self.ax)
        self.ax.set_visible(True)
        self.ax_geo.set_visible(False)
        # plt.plot([1, 2, 3], [1, 2, 3])
        self.ax.figure.canvas.draw()

    # Define function for switching to ax2
    def switch_to_ax_geo(self, event):
        """Switch to geographic axes."""
        self.fig.canvas.manager.toolbar.home()
        plt.sca(self.ax_geo)
        self.ax.set_visible(False)
        self.ax_geo.set_visible(True)
        # plt.plot([1, 2, 3], [3, 2, 1])
        self.ax_geo.figure.canvas.draw()


class AoiSelect(BaseSelect):
    """Selector tool to provide source GCP coordinates to pyOpenRiverCam."""

    def __init__(self, img, src, dst, camera_config, logger=logging):
        """Set up interactive plot for defining region of interest (AOI)."""
        if hasattr(camera_config, "crs"):
            crs = camera_config.crs
        else:
            crs = None
        super(AoiSelect, self).__init__(img, dst, crs=crs, logger=logger)
        # make empty plot
        self.camera_config = camera_config
        (self.p_gcps,) = self.ax.plot(
            *list(zip(*src)), "o", color="w", markeredgecolor="k", markersize=10, zorder=3, label="GCPs"
        )
        self.pts_t_gcps = [
            self.ax.annotate(
                n + 1,
                xytext=(6, 6),
                xy=xy,
                textcoords="offset points",
                zorder=4,
                path_effects=[
                    patheffects.Stroke(linewidth=3, foreground="w"),
                    patheffects.Normal(),
                ],
            )
            for n, xy in enumerate(src)
        ]
        # self.pts_t.append(pt)

        (self.p,) = self.ax.plot([], [], "o", markersize=10, color="c", markeredgecolor="w", zorder=3)
        kwargs = dict(
            markersize=10,
            color="c",
            markeredgecolor="w",
            zorder=3,
        )
        if hasattr(self.camera_config, "crs") and use_cartopy:
            kwargs["transform"] = ccrs.PlateCarree()
        (self.p_geo,) = self.ax_geo.plot([], [], "o", **kwargs)
        # plot an empty polygon
        pol = Polygon(np.zeros((0, 2)), edgecolor="w", alpha=0.5, linewidth=2)
        if hasattr(self.camera_config, "crs") and use_cartopy:
            pol_geo = Polygon(
                np.zeros((0, 2)), edgecolor="w", alpha=0.5, linewidth=2, transform=ccrs.PlateCarree(), zorder=3
            )
        else:
            pol_geo = Polygon(np.zeros((0, 2)), edgecolor="w", alpha=0.5, linewidth=2, zorder=3)
        self.p_bbox_geo = self.ax_geo.add_patch(pol_geo)
        self.p_bbox = self.ax.add_patch(pol)
        xloc = self.ax.get_xlim()[0] + 50
        yloc = self.ax.get_ylim()[-1] + 50
        self.title = self.ax.text(
            xloc,
            yloc,
            "Select corners of bounding box in the following order with left and right as seen looking "
            "downstream:\nUpstream-left\nDownstream-left\nDownstream-right\nUpstream-right",
            size=12,
            va="top",
            path_effects=path_effects,
        )

        # # TODO: if no crs provided, then provide a normal axes with equal lengths on x and y axis
        self.required_clicks = 4
        # self.cam_config_cam = None
        # self.cam_config_geo = None
        plt.show(block=True)

    def on_left_click(self, event):
        """Define left-click event."""
        if event.xdata is not None:
            self.logger.debug(f"Storing coordinate x: {event.xdata} y: {event.ydata} to src")
            self.src.append([int(np.round(event.xdata)), int(np.round(event.ydata))])
            self.p.set_data(*list(zip(*self.src)))
            pt = self.ax.annotate(
                corner_labels[len(self.src) - 1],
                xytext=(6, 6),
                xy=(event.xdata, event.ydata),
                textcoords="offset points",
                zorder=4,
                path_effects=[
                    patheffects.Stroke(linewidth=3, foreground="w"),
                    patheffects.Normal(),
                ],
            )
            self.pts_t.append(pt)
            # check if all points are complete
            if len(self.src) == self.required_clicks:
                try:
                    self.camera_config.set_bbox_from_corners(self.src)
                    bbox_cam = list(
                        zip(
                            *self.camera_config.get_bbox(
                                mode="camera", expand_exterior=True, within_image=True
                            ).exterior.xy,
                        )
                    )
                    bbox_geo = list(
                        zip(
                            *self.camera_config.get_bbox(expand_exterior=False, within_image=True).exterior.xy,
                        )
                    )
                    if hasattr(self.camera_config, "crs") and use_cartopy:
                        bbox_geo = helpers.xyz_transform(bbox_geo, crs_from=self.camera_config.crs, crs_to=4326)
                    self.p_bbox.set_xy(bbox_cam)
                    self.p_bbox_geo.set_xy(bbox_geo)
                    self.ax.figure.canvas.draw()
                except Exception:
                    self.title.set_text(
                        "Could not resolve bounding box with the set coordinates.\nThe coordinates are likely "
                        "measured with too low accuracy.\nMeasure with cm accuracy."
                    )

    def on_click(self, event):
        """Define general click event."""
        super(AoiSelect, self).on_click(event)
        if not (len(self.src) == self.required_clicks):
            # remove plot if present
            self.p_bbox.set_xy(np.zeros((0, 2)))
            self.p_bbox_geo.set_xy(np.zeros((0, 2)))
            self.ax.figure.canvas.draw()


class GcpSelect(BaseSelect):
    """Selector tool to provide source GCP coordinates to pyOpenRiverCam."""

    def __init__(self, img, dst, crs=None, camera_matrix=None, dist_coeffs=None, lens_position=None, logger=logging):
        """Set up interactive GCP selection plot."""
        super(GcpSelect, self).__init__(img, dst, crs=crs, logger=logger)
        # make empty plot
        (self.p,) = self.ax.plot(
            [], [], "o", color="w", markeredgecolor="k", markersize=10, zorder=3, label="Clicked control points"
        )
        kwargs = dict(color="c", markeredgecolor="w", zorder=4, markersize=10, label="Selected control points")
        if crs is not None and use_cartopy:
            kwargs["transform"] = ccrs.PlateCarree()
        (self.p_geo_selected,) = self.ax_geo.plot([], [], "o", **kwargs)
        if len(dst) >= 4:
            # plot an empty set of crosses for the fitted gcp row columns after optimization of perspective
            (self.p_fit,) = self.ax.plot([], [], "+", markersize=10, color="r", zorder=4, label="Fitted control_points")
        else:
            self.p_fit = None

        xloc = self.ax.get_xlim()[0] + 50
        yloc = self.ax.get_ylim()[-1] + 50
        self.title = self.ax.text(
            xloc,
            yloc,
            "Select location of control points in the right order (check locations on map view)",
            size=12,
            path_effects=path_effects,
        )
        self.ax_geo.legend()
        self.ax.legend()
        self.lens_position = lens_position
        self.camera_matrix_input = camera_matrix
        self.dist_coeffs_input = dist_coeffs
        # add dst coords in the intended CRS
        if crs is not None and use_cartopy:
            self.dst_crs = helpers.xyz_transform(self.dst, 4326, crs)
        else:
            self.dst_crs = self.dst
        self.required_clicks = len(self.dst)
        self.camera_matrix = None
        self.dist_coeffs = None

    def on_left_click(self, event):
        """Define left-click event."""
        super(GcpSelect, self).on_left_click(event)
        # figure out if the fitted control points must be computed and plotted
        if self.p_fit is not None:
            if len(self.src) == self.required_clicks:
                self.title.set_text("Fitting pose and camera parameters...")
                self.ax.figure.canvas.draw()
                src_fit, dst_fit, camera_matrix, dist_coeffs, rvec, tvec, err = cli_utils.get_gcps_optimized_fit(
                    self.src,
                    self.dst_crs,
                    self.height,
                    self.width,
                    c=2.0,
                    camera_matrix=self.camera_matrix_input,
                    dist_coeffs=self.dist_coeffs_input,
                    lens_position=self.lens_position,
                )
                self.p_fit.set_data(*list(zip(*src_fit)))
                self.camera_matrix = camera_matrix
                self.dist_coeffs = dist_coeffs
                new_text = (
                    'Pose and camera parameters fitted (see "+"), average x, y distance error: {:1.3f} m.'.format(err)
                )
                if err > 0.1:
                    new_text += (
                        "\nWarning: error is larger than 0.1 meter. Are you sure that the coordinates are "
                        "measured accurately?"
                    )
                self.title.set_text(new_text)
                self.ax.figure.canvas.draw()
            else:
                self.p_fit.set_data([], [])

    def on_right_click(self, event):
        """Define right-click event."""
        super(GcpSelect, self).on_right_click(event)
        if self.p_fit is not None:
            if len(self.src) < self.required_clicks:
                self.p_fit.set_data([], [])

    def on_click(self, event):
        """Define general click event."""
        super(GcpSelect, self).on_click(event)
        # update selected dst points
        dst_sel = self.dst[: len(self.src)]
        if len(dst_sel) > 0:
            self.p_geo_selected.set_data(*list(zip(*dst_sel))[0:2])
        else:
            self.p_geo_selected.set_data([], [])


class StabilizeSelect(BaseSelect):
    """Stabilization plot."""

    def __init__(self, img, logger=logging):
        """Set up stabilization interactive plot."""
        super(StabilizeSelect, self).__init__(img, logger=logger)
        # make empty plot
        pol = Polygon(np.zeros((0, 2)), edgecolor="w", alpha=0.5, linewidth=2)
        self.p = self.ax.add_patch(pol)
        # kwargs = dict(color="c", markeredgecolor="w", zorder=4, markersize=10, label="Selected control points")
        xloc = self.ax.get_xlim()[0] + 50
        yloc = self.ax.get_ylim()[-1] + 50
        self.title = self.ax.text(
            xloc,
            yloc,
            "Select a polygon of at least 4 points that encompasses at least the water surface. Areas outside will be "
            "treated as stable areas for stabilization.",
            size=12,
            path_effects=path_effects,
        )
        # self.ax.legend()
        # add dst coords in the intended CRS
        self.required_clicks = 4  # minimum 4 points needed for a satisfactory ROI

    def on_click(self, event):
        """Define general click event."""
        # only if ax is visible and click was within the window
        if self.ax.get_visible() and event.inaxes == self.ax:
            if event.button is MouseButton.RIGHT:
                self.on_right_click(event)
            elif event.button is MouseButton.LEFT:
                self.on_left_click(event)
                # check if enough points are collected to enable the Done button
            if len(self.src) >= self.required_clicks:
                self.button3.set_active(True)
                self.button3.label.set_color("k")
            else:
                self.button3.set_active(False)
                self.button3.label.set_color("grey")

        self.ax.figure.canvas.draw()

    def on_right_click(self, event):
        """Define right-click event."""
        if len(self.pts_t) > 0:
            self.pts_t[-1].remove()
            del self.pts_t[-1]
        if len(self.src) > 0:
            del self.src[-1]
            if len(self.src) > 0:
                self.p.set_xy(self.src)
            else:
                self.p.set_xy(np.zeros((0, 2)))
        self.ax.figure.canvas.draw()

    def on_left_click(self, event):
        """Define left-click event."""
        if event.xdata is not None:
            self.logger.debug(f"Storing coordinate x: {event.xdata} y: {event.ydata} to src")
            self.src.append([int(np.round(event.xdata)), int(np.round(event.ydata))])
            # self.p.set_data(*list(zip(*self.src)))
            self.p.set_xy(self.src)
            pt = self.ax.annotate(
                len(self.src),
                xytext=(6, 6),
                xy=(event.xdata, event.ydata),
                textcoords="offset points",
                zorder=4,
                path_effects=[
                    patheffects.Stroke(linewidth=3, foreground="w"),
                    patheffects.Normal(),
                ],
            )
            self.pts_t.append(pt)
            self.ax.figure.canvas.draw()

    def on_close(self, event):
        """Define on-close event handler."""
        # overrule the amount of required clicks
        self.required_clicks = len(self.src)
        super(StabilizeSelect, self).on_close(event)



"""WaterLevel module for pyorc."""

from __future__ import annotations

import warnings
from typing import Dict, List, Literal, Optional, Tuple, Union

import geopandas as gpd
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patheffects
from scipy.interpolate import interp1d
from scipy.optimize import differential_evolution
from shapely import affinity, geometry
from shapely.ops import polygonize

from pyorc import cv, plot_helpers

MODES = Literal["camera", "geographic", "cross_section"]
PATH_EFFECTS = [
    patheffects.Stroke(linewidth=3, foreground="w"),
    patheffects.Normal(),
]
LINE_COLOR = "#385895"
PLANAR_KWARGS = {"color": "c", "alpha": 0.5}
BOTTOM_KWARGS = {"color": "brown", "alpha": 0.1}
BANK_OPTIONS = {"far", "near", "both"}
WETTED_KWARGS = {
    "alpha": 0.15,
    "linewidth": 2.0,
    "facecolor": LINE_COLOR,
    "path_effects": PATH_EFFECTS,
    "edgecolor": "w",
    "zorder": 1,
}


def _make_angle_lines(csl_points, angle_perp, length, offset):
    """Make lines at cross section points, perpendicular to cross section orientation using angle and offset."""
    # move points
    csl_points = [
        affinity.translate(p, xoff=np.cos(angle_perp) * offset, yoff=np.sin(angle_perp) * offset) for p in csl_points
    ]

    # make horizontally oriented lines of the required length (these are only xy at this stage)
    csl_lines = [
        geometry.LineString([geometry.Point(p.x - length / 2, p.y), geometry.Point(p.x + length / 2, p.y)])
        for p in csl_points
    ]
    # rotate in counter-clockwise perpendicular direction to the orientation of the cross section itself
    csl_lines = [affinity.rotate(l, angle_perp, origin=p, use_radians=True) for l, p in zip(csl_lines, csl_points)]
    return csl_lines


def _histogram(data, bin_size: int = 5, normalize=False):
    """Create a histogram with predefined bin size."""
    bin_size = int(bin_size)
    if not data.dtype == np.uint8:
        raise ValueError("Image data must be of type uint8.")
    if not (bin_size >= 5 and bin_size <= 20):
        raise ValueError("Bin size must be between 5 and 20")
    bins = np.arange(0, 256, bin_size)
    # Compute histogram counts
    counts, edges = np.histogram(data, bins=bins)

    # Normalize counts if required
    if normalize and np.sum(counts) > 0:
        bin_widths = np.diff(edges)
        counts = counts / (np.sum(counts) * bin_widths)  # Normalize to sum to 1 over the distribution

    # Calculate bin centers
    centers = (edges[:-1] + edges[1:]) / 2

    return centers, edges, counts


def _histogram_union(edges, hist1, hist2):
    """Measures the union of two normalized histograms and turns into score. They must have the same bins.

    A score of 0 means the histograms are entirely different, score of 1 means they are identical.
    """
    bin_chunks = edges[1:] - edges[:-1]
    # take the piecewise maximum of histograms
    hist_max = np.maximum(hist1, hist2)
    # integrate
    union = (bin_chunks * hist_max).sum()
    # normalize score: 1 least optimal, 0 most optimal.
    return 2 - union


class CrossSection:
    """Water Level functionality."""

    def __str__(self):
        return str(self.cs_linestring)

    def __repr__(self):
        return str(self.cs_linestring)

    def __init__(self, camera_config, cross_section: Union[gpd.GeoDataFrame, List[List]]):
        """Initialize a cross-section representation.

        The object is geographically aware through the camera configuration.
        If a GeoDataFrame is supplied as `cross_section` input, CRS consistency
        between the cross-section and camera configuration is guaranteed.

        Parameters
        ----------
        camera_config : object
            Defines the camera configuration, including potential CRS used for processing
            geographic data consistency.
        cross_section : GeoDataFrame or list of lists
            Cross-sectional data representing coordinate information. Can either be a GeoDataFrame
            with CRS potentially set or a list-of-lists where each inner list contains [x, y, z] coordinates.

        Attributes
        ----------
        x : list of float
            Represents the x-coordinates of the cross-section. For GeoDataFrame input, these are
            converted into the CRS of the camera configuration if applicable.
        y : list of float
            Represents the y-coordinates of the cross-section. For GeoDataFrame input, these are
            converted into the CRS of the camera configuration if applicable.
        z : list of float
            Represents the z-coordinates of the cross-section.
        s : numpy.ndarray
            Represents the horizontal cumulative distances calculated between consecutive points
            in the cross-section, used primarily for spatial interpolation.
        l : numpy.ndarray
            Represents the cumulative spatial coordinates, tracking both horizontal and vertical
            distances over the cross-section, from the left-most to the right-most section.
            used primarily for spatial interpolation and needed where pure horizontal or vertical lines are experienced.
        camera_config : object
            Associated camera configuration object indicating CRS and related configuration applied
            to handle projection and compatibility with cross-section data.

        Raises
        ------
        ValueError
            Raised if the CRS is defined for only one of `camera_config` or `cross_section`,
            as CRS for both or neither must be specified for consistency.
        TypeError
            Raised if `cross_section` is neither of type GeoDataFrame nor a list of lists structure.

        """
        # if cross_section is a GeoDataFrame, check if it has a CRS, if yes, convert coordinates to crs of CameraConfig
        if isinstance(cross_section, gpd.GeoDataFrame):
            crs_cs = getattr(cross_section, "crs", None)
            crs_cam = getattr(camera_config, "crs", None)
            if crs_cs is not None and crs_cam is not None:
                cross_section.to_crs(camera_config.crs, inplace=True)
            elif crs_cs is not None or crs_cam is not None:
                raise ValueError("if a CRS is used, then both camera_config and cross_section must have a CRS.")
            g = cross_section.geometry
            x, y, z = g.x.values, g.y.values, g.z.values
        else:
            x, y, z = list(map(list, zip(*cross_section)))

        x_diff = np.concatenate((np.array([0]), np.diff(x)))
        y_diff = np.concatenate((np.array([0]), np.diff(y)))
        z_diff = np.concatenate((np.array([0]), np.diff(z)))
        # estimate distance from left to right bank
        s = np.cumsum((x_diff**2 + y_diff**2) ** 0.5)
        lens_position_xy = camera_config.estimate_lens_position()[0:2]
        d = ((x - lens_position_xy[0]) ** 2 + (y - lens_position_xy[1]) ** 2) ** 0.5
        # estimate length coordinates
        l = np.cumsum(np.sqrt(x_diff**2 + y_diff**2 + z_diff**2))

        self.x = np.array(x)  # x-coordinates in local projection or crs
        self.y = np.array(y)  # y-coordinates in local projection or crs
        self.z = np.array(z)  # z-coordinates in local projection or crs
        self.s = s  # horizontal distance from left to right bank (only used for interpolation funcs).
        self.l = l  # length distance (following horizontal and vertical) over cross section from left to right
        self.d = d  # horizontal distance between lens horizontal position and cross section point
        self.camera_config = camera_config

    @property
    def interp_x(self) -> interp1d:
        """Linear interpolation function for x-coordinates, using l as input."""
        return interp1d(self.l, self.x, kind="linear", fill_value="extrapolate")

    @property
    def interp_y(self) -> interp1d:
        """Linear interpolation function for y-coordinates, using l as input."""
        return interp1d(self.l, self.y, kind="linear", fill_value="extrapolate")

    @property
    def interp_z(self) -> interp1d:
        """Linear interpolation function for z-coordinates, using l as input."""
        return interp1d(self.l, self.z, kind="linear", fill_value="extrapolate")

    @property
    def interp_d(self) -> interp1d:
        """Linear interpolation function for distances (d) to lens, using l as input."""
        return interp1d(self.l, self.d, kind="linear", fill_value="extrapolate")

    @property
    def interp_x_from_s(self) -> interp1d:
        """Linear interpolation function for x-coordinates, using s as input."""
        return interp1d(self.s, self.x, kind="linear", fill_value="extrapolate")

    @property
    def interp_y_from_s(self) -> interp1d:
        """Linear interpolation function for y-coordinates, using s as input.."""
        return interp1d(self.s, self.y, kind="linear", fill_value="extrapolate")

    @property
    def interp_z_from_s(self) -> interp1d:
        """Linear interpolation function for z-coordinates, using s as input.."""
        return interp1d(self.s, self.z, kind="linear", fill_value="extrapolate")

    @property
    def interp_s_from_l(self) -> interp1d:
        """Linear interpolation function for s-coordinates, from left to right bank, using l as input."""
        return interp1d(self.l, self.s, kind="linear", fill_value="extrapolate")

    @property
    def cs_points(self) -> List[geometry.Point]:
        """Return cross-section as list of shapely.geometry.Point."""
        return [geometry.Point(_x, _y, _z) for _x, _y, _z in zip(self.x, self.y, self.z)]

    @property
    def cs_points_sz(self) -> List[geometry.Point]:
        """Return cross-section perpendicular to flow direction (SZ) as list of shapely.geometry.Point."""
        return [geometry.Point(_s, _z) for _s, _z in zip(self.s, self.z)]

    @property
    def cs_linestring(self) -> geometry.LineString:
        """Return cross-section as shapely.geometry.Linestring."""
        return geometry.LineString(self.cs_points)

    @property
    def cs_linestring_sz(self) -> geometry.LineString:
        """Return cross-section perpendicular to flow direction (SZ) as shapely.geometry.Linestring."""
        return geometry.LineString(self.cs_points_sz)

    @property
    def cs_angle(self):
        """Average angle orientation that cross-section makes in geographical space.

        Zero means left to right direction, positive means counter-clockwise, negative means clockwise.
        """
        point1_xy = self.cs_points[0].coords[0][:-1]
        point2_xy = self.cs_points[-1].coords[0][:-1]
        diff_xy = np.array(point2_xy) - np.array(point1_xy)
        return np.arctan2(diff_xy[1], diff_xy[0])

    @property
    def distance_camera(self):
        """Estimate distance of mean coordinate of cross section to camera position."""
        coord_mean = np.mean(self.cs_linestring.coords, axis=0)
        return np.sum((self.camera_config.estimate_lens_position() - coord_mean) ** 2) ** 0.5

    @property
    def idx_closest_point(self):
        """Determine index of point in cross-section, closest to the camera."""
        return 0 if self.d[0] < self.d[-1] else len(self.d) - 1

    @property
    def idx_farthest_point(self):
        """Determine index of point in cross-section, farthest from the camera."""
        return 0 if self.d[0] > self.d[-1] else len(self.d) - 1

    @property
    def within_image(self):
        """Check if any of the points of the cross section fall inside the image objective."""
        # check if cross section is visible within the image objective
        pix = self.camera_config.project_points(np.array(list(map(list, self.cs_linestring.coords))), within_image=True)
        # check which points fall within the image objective
        within_image = np.all([pix[:, 0] >= 0, pix[:, 0] < 1920, pix[:, 1] >= 0, pix[:, 1] < 1080], axis=0)
        # check if there are any points within the image objective and return result
        return bool(np.any(within_image))

    def get_cs_waterlevel(self, h: float, sz=False, extend_by=None) -> geometry.LineString:
        """Retrieve LineString of water surface at cross-section at a given water level.

        Parameters
        ----------
        h : float
            water level [m]
        sz : bool, optional
            If set, return water level line in y-z projection, by default False.
        extend_by : float, optional
            If set, the line will be extended left and right using the defined float in meters

        Returns
        -------
        geometry.LineString
            horizontal line at water level (2d if `sz`=True, 3d if `yz`=False)

        """
        # get water level in camera config vertical datum
        z = self.camera_config.h_to_z(h)
        if sz:
            if extend_by is None:
                s_coords = self.s
            else:
                s_coords = np.concatenate([[-np.abs(extend_by)], self.s, [self.s[-1] + np.abs(extend_by)]])
            return geometry.LineString(zip(s_coords, [z] * len(s_coords)))
        if extend_by is not None:
            alpha = np.arctan((self.x[1] - self.x[0]) / (self.y[1] - self.y[0]))
            x_coords = np.concatenate(
                [
                    [self.x[0] - np.cos(alpha) * np.abs(extend_by)],
                    self.x,
                    [self.x[-1] + np.cos(alpha) * np.abs(extend_by)],
                ]
            )
            y_coords = np.concatenate(
                [
                    [self.y[0] - np.sin(alpha) * np.abs(extend_by)],
                    self.y,
                    [self.y[-1] + np.sin(alpha) * np.abs(extend_by)],
                ]
            )
        else:
            x_coords = self.x
            y_coords = self.y

        return geometry.LineString(zip(x_coords, y_coords, [z] * len(x_coords)))

    def get_csl_point(
        self, h: Optional[float] = None, l: Optional[float] = None, camera: bool = False, swap_y_coords: bool = False
    ) -> List[geometry.Point]:
        """Retrieve list of points, where cross-section (cs) touches the land (l).

        Multiple points may be found.

        Parameters
        ----------
        h : float, optional
            water level [m], if provided, s must not be provided.
        l : float
            coordinate of distance from left to right bank, including height [m], if provided, h must not be provided.
        camera : bool, optional
            If set, return 2D projected points, by default False.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)

        Returns
        -------
        List[shapely.geometry.Point]
            List of points, where water line touches land, can be only one or two points.

        """
        if h is not None and l is not None:
            raise ValueError("Only one of h or l can be provided.")
        if h is None and l is None:
            raise ValueError("One of h or l must be provided.")
        # get water level in camera config vertical datum
        if l is not None:
            if l < 0 or l > self.l[-1]:
                raise ValueError(
                    "Value of l is lower (higher) than the minimum (maximum) value found in the cross section"
                )
            cross = [geometry.Point(self.interp_x(l), self.interp_y(l), self.interp_z(l))]
        else:
            z = self.camera_config.h_to_z(h)
            if z > np.array(self.z).max() or z < np.array(self.z).min():
                raise ValueError(
                    "Value of water level is lower (higher) than the minimum (maximum) value found in the cross section"
                )
            cs_waterlevel = self.get_cs_waterlevel(h, sz=True)
            # get crossings and turn into list of geometry.Point
            cross_sz = cs_waterlevel.intersection(self.cs_linestring_sz)

            # make cross_yz iterable
            if isinstance(cross_sz, geometry.Point):
                cross_sz = [cross_sz]
            elif isinstance(cross_sz, geometry.MultiPoint):
                cross_sz = list(cross_sz.geoms)
            else:
                raise ValueError(
                    "Cross section is not crossed by water level. Check if water level is within the cross section."
                )
            # find xyz real-world coordinates and turn in to POINT Z list
            cross = [
                geometry.Point(self.interp_x_from_s(c.xy[0]), self.interp_y_from_s(c.xy[0]), c.xy[1]) for c in cross_sz
            ]

        if camera:
            coords = [[p.x, p.y, p.z] for p in cross]
            coords_proj = self.camera_config.project_points(coords, swap_y_coords=swap_y_coords)
            cross = [geometry.Point(p[0], p[1]) for p in coords_proj]
        return cross

    def get_csl_line(
        self,
        h: Optional[float] = None,
        l: Optional[float] = None,
        length: float = 0.5,
        offset: float = 0.0,
        camera: bool = False,
        swap_y_coords: bool = False,
    ) -> List[geometry.LineString]:
        """Retrieve waterlines over the cross-section, perpendicular to the orientation of the cross-section.

        Returns a 2D LineString if camera is True, 3D if False

        Parameters
        ----------
        h : float, optional
            water level [m]
        l : float
            coordinate of distance from left to right bank, including height [m], if provided, h must not be provided.
        length : float, optional
            length of the waterline [m], by default 0.5
        offset : float, optional
            perpendicular offset of the waterline from the cross-section [m], by default 0.0
        camera : bool, optional
            If set, return 2D projected lines, by default False.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)

        Returns
        -------
        List[shapely.geometry.LineString]
            List of lines perpendicular to cross section orientation, can be only one or two lines.

        """
        csl_points = self.get_csl_point(h=h, l=l)
        z = csl_points[0].z

        # acquire angle perpendicular to cross section
        angle_perp = self.cs_angle + np.pi / 2

        # retrieve lines perpendicular to cross section with offset and length
        csl_lines = _make_angle_lines(csl_points, angle_perp, length, offset)

        if camera:
            # transform to 2D projected lines, make list of lists of coordinates
            coords_lines = [[[_x, _y, z] for _x, _y in l.coords] for l in csl_lines]
            # project list of lists
            coords_lines_proj = [
                self.camera_config.project_points(cl, swap_y_coords=swap_y_coords) for cl in coords_lines
            ]
            # turn list of lists into list of LineString
            csl_lines = [geometry.LineString([geometry.Point(_x, _y) for _x, _y in l]) for l in coords_lines_proj]
        else:
            # add a z-coordinate and return
            csl_lines = [geometry.LineString([geometry.Point(_x, _y, z) for _x, _y in l.coords]) for l in csl_lines]
        return csl_lines

    def get_csl_pol(
        self,
        h: Optional[float] = None,
        l: Optional[float] = None,
        length: float = 0.5,
        padding: Tuple[float, float] = (-0.5, 0.5),
        offset: float = 0.0,
        camera: bool = False,
        swap_y_coords: bool = False,
    ) -> List[geometry.Polygon]:
        """Get horizontal polygon from cross-section land-line towards water or towards land.

        Returns a 2D Polygon if camera is True, 3D if False

        Parameters
        ----------
        h : float
            water level [m]
        l : float
            coordinate of distance from left to right bank, including height [m], if provided, h must not be provided.
        length : float, optional
            length of the waterline [m], by default 0.5
        padding : Tuple[float, float], optional
            amount of distance [m] to extend the polygon beyond the waterline, by default (-0.5, 0.5)
        offset : float, optional
            perpendicular offset of the waterline from the cross-section [m], by default 0.0
        camera : bool, optional
            If set, return 2D projected polygons, by default False.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)

        Returns
        -------
        List[shapely.geometry.Polygon]
            List of lines perpendicular to cross section orientation, can be only one or two lines.


        """
        # retrieve water line(s)
        csl = self.get_csl_line(h=h, l=l, length=length, offset=offset)
        if len(padding) != 2:
            raise ValueError(f"padding must contain two values (provided: {len(padding)}")
        if padding[1] <= padding[0]:
            raise ValueError("First value of padding must be smaller than second")
        csl_pol_bounds = [
            [
                affinity.translate(
                    line, xoff=np.cos(self.cs_angle) * padding[0], yoff=np.sin(self.cs_angle) * padding[0]
                ),
                affinity.translate(
                    line, xoff=np.cos(self.cs_angle) * padding[1], yoff=np.sin(self.cs_angle) * padding[1]
                ),
            ]
            for line in csl
        ]
        csl_pol_coords = [
            list(lines[0].coords) + list(lines[1].coords[::-1]) + [lines[0].coords[0]] for lines in csl_pol_bounds
        ]
        if camera:
            coords_expand = np.zeros((0, 3))
            for cn, coords in enumerate(csl_pol_coords):
                for n in range(0, len(coords) - 1):
                    new_coords = np.linspace(coords[n], coords[n + 1], 100)
                    coords_expand = np.r_[coords_expand, new_coords]
                coords = coords_expand
                csl_pol_coords[cn] = coords
            csl_pol_coords = [
                self.camera_config.project_points(coords, swap_y_coords=swap_y_coords, within_image=True)
                for coords in csl_pol_coords
            ]
            csl_pol_coords = [coords[np.isfinite(coords[:, 0])] for coords in csl_pol_coords]
        polygons = [geometry.Polygon(coords) for coords in csl_pol_coords]
        return polygons

    def get_bottom_surface(self, length: float = 2.0, offset: float = 0.0, camera: bool = False, swap_y_coords=False):
        """Retrieve a bottom surface polygon for the entire cross section, expanded over a length.

        Returns a 2D Polygon if camera is True, 3D if False. Useful in particular for 3d situation plots.

        Parameters
        ----------
        length : float, optional
            length of the cross-section lateral expansion (from up to downstream) [m], by default 2.0.
        offset : float, optional
            perpendicular offset of the expanded cross section from the original cross section coordinates [m],
            by default 0.0.
        camera : bool, optional
            If set, return 2D projected polygon, by default False. Note that any plotting does not provide shading or
            ray tracing hence plotting a 2D polygon is not recommended.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)

        Returns
        -------
        shapely.geometry.Polygon
            polygon representing the cross-section expanded over a certain length (2d if camera=True,
            3d if camera=False).

        See Also
        --------
        CrossSection.plot_bottom_surface : Plot the bottom surface resulting from this method.

        """
        csl_points = [self.cs_points[0], self.cs_points[-1]]
        angle_perp = self.cs_angle + np.pi / 2
        # retrieve lines perpendicular to cross section with offset and length
        csl_lines = _make_angle_lines(csl_points, self.cs_angle + np.pi / 2, length, offset)
        csl_line_points = [
            [geometry.Point(_x, _y, z) for _x, _y in l.coords]
            for l, z in zip(csl_lines, [self.cs_points[0].z, self.cs_points[-1].z])
        ]
        # retrieve lines of displaced cross sections
        csl_displaced = [
            [
                affinity.translate(p, xoff=np.cos(angle_perp) * (offset + l), yoff=np.sin(angle_perp) * (offset + l))
                for p in self.cs_points
            ]
            for l in [length / 2, -length / 2]
        ]
        # put together a full polygon
        all_points = csl_line_points[0] + csl_displaced[0] + csl_line_points[1][::-1] + csl_displaced[1][::-1]
        if camera:
            coords = np.array([list(p.coords[0]) for p in all_points])

            all_points = self.camera_config.project_points(coords, swap_y_coords=swap_y_coords, within_image=True)
            all_points = all_points[np.isfinite(all_points[:, 0])]
            all_points = [geometry.Point(*p) for p in all_points]

        return geometry.Polygon(all_points)

    def get_planar_surface(
        self, h: float, length: float = 2.0, offset: float = 0.0, camera: bool = False, swap_y_coords=False
    ) -> geometry.Polygon:
        """Retrieve a planar water surface for a given water level, as a geometry.Polygon.

        Returns a 2D Polygon if camera is True, 3D if False

        Parameters
        ----------
        h : float
            water level [m]
        length : float, optional
            length of the waterline [m], by default 2.0
        offset : float, optional
            perpendicular offset of the waterline from the cross section [m], by default 0.0
        camera : bool, optional
            If set, return 2D projected polygon, by default False.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)

        Returns
        -------
        shapely.geometry.Polygon
            rectangular horizontal polygon representing the planar water surface (2d if camera=True,
            3d if camera=False).

        See Also
        --------
        CrossSection.plot_planar_surface : Plot the planar surface resulting from this method.

        """
        wls = self.get_csl_line(h=h, offset=offset, length=length, camera=camera, swap_y_coords=swap_y_coords)
        if len(wls) != 2:
            raise ValueError("Amount of water line crossings must be 2 for a planar surface estimate.")
        return geometry.Polygon(list(wls[0].coords) + list(wls[1].coords[::-1]))

    def get_wetted_surface_sz(self, h: float) -> geometry.Polygon:
        """Retrieve a wetted surface perpendicular to flow direction (SZ) for a water level, as a geometry.Polygon.

        This is a useful method for instance to estimate m2 wetted surface for a given water level in the cross
        section.

        Parameters
        ----------
        h : float
            water level [m]

        Returns
        -------
        geometry.Polygon
            Wetted surface as a polygon, in Y-Z projection.

        """
        wl = self.get_cs_waterlevel(
            h=h, sz=True, extend_by=0.1
        )  # extend a small bit to guarantee crossing with the bottom coordinates
        zl = wl.xy[1][0]
        # create polygon by making a union
        bottom_points = self.cs_points_sz
        # add a point left and right slightly above the level if the level is below the water level
        if bottom_points[0].y < zl:
            bottom_points.insert(0, geometry.Point(bottom_points[0].x, zl + 0.1))
        if bottom_points[-1].y < zl:
            bottom_points.append(geometry.Point(bottom_points[-1].x, zl + 0.1))
        bottom_line = geometry.LineString(bottom_points)
        pol = list(polygonize(wl.union(bottom_line)))
        if len(pol) == 0:
            # create infinitely small polygon at lowest z coordinate
            lowest_z = min(self.z)
            lowest_s = self.s[list(self.z).index(lowest_z)]
            # make an infinitely small polygon around the lowest point in the cross section
            pol = [geometry.Polygon([(lowest_s, lowest_z)] * 3)]
        elif len(pol) > 1:
            # detect which polygons have their average z coordinate below the defined water level
            pol = [p for p in pol if p.centroid.xy[1][0] < zl]
            # raise ValueError("Water level is crossed by multiple polygons.")
        # else:
        #     pol = pol[0]
        return geometry.MultiPolygon(pol)

    def get_wetted_surface(self, h: float, camera: bool = False, swap_y_coords=False) -> geometry.Polygon:
        """Retrieve a wetted surface for a given water level, as a geometry.Polygon.

        Parameters
        ----------
        h : float
            water level [m]
        camera : bool, optional
            If set, return 2D projected polygon, by default False.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)

        Returns
        -------
        geometry.Polygon
            Wetted surface as a polygon (2d if camera=True, 3d if camera=False).


        """
        pols = self.get_wetted_surface_sz(h=h)
        pols_proj = []
        for pol in pols.geoms:
            coords = [[self.interp_x_from_s(p[0]), self.interp_y_from_s(p[0]), p[1]] for p in pol.exterior.coords]
            if camera:
                coords_proj = self.camera_config.project_points(coords, swap_y_coords=swap_y_coords)
                pols_proj.append(geometry.Polygon(coords_proj))
            else:
                pols_proj.append(geometry.Polygon(coords))
        return geometry.MultiPolygon(pols_proj)

    def get_line_of_interest(self, bank: BANK_OPTIONS = "far") -> List[float]:
        """Retrieve the points of interest within the cross-section for water level detection.

        This may be all points, points only at the far-bank or closest-bank of the camera position.

        Parameters
        ----------
        bank: Literal["far", "near", "both"], optional
            Select relevant part of cross section. If "both", select the full cross-section. If "far", select
            only the furthest part from the deepest point in the cross section, if any. If "near", select only
            the "nearest".

        Returns
        -------
        list of 2 floats
            start and end point of the line of interest in l-coordinates.

        """
        if bank == "both":
            return self.l.min(), self.l.max()
        elif bank == "far":
            start_point = self.l[self.idx_farthest_point]

        elif bank == "near":
            start_point = self.l[self.idx_closest_point]
        else:
            raise ValueError(f"bank must be one of {BANK_OPTIONS}, not {bank}")
        # find l values at lowest point

        l_lowest = self.l[np.where(self.z == self.z.min())]
        # find which is closest to start_point
        end_point = l_lowest[np.argmin(np.abs(l_lowest - start_point))]

        # sort from low to high
        return tuple(np.sort(np.array([start_point, end_point])))
        # fin l-value closest to

    def get_histogram_score(
        self,
        x: float,
        img: np.array,
        bin_size: float = 5.0,
        offset: float = 0.0,
        padding: float = 0.5,
        length: float = 2.0,
        min_z: Optional[float] = None,
        max_z: Optional[float] = None,
        min_samples: int = 50,
    ):
        """Retrieve a histogram score for a given l-value."""
        l = x[0]
        if min_z is not None:
            if self.interp_z(l) < min_z:
                # return worst score
                return 2.0 + np.abs(self.interp_z(l) - min_z)
        elif max_z is not None:
            if self.interp_z(l) > max_z:
                return 2.0 + np.abs(self.interp_z(l) - max_z)
        pol1 = self.get_csl_pol(l=l, offset=offset, padding=(0, padding), length=length, camera=True)[0]
        pol2 = self.get_csl_pol(l=l, offset=offset, padding=(-padding, 0), length=length, camera=True)[0]
        # get intensity values within polygons
        ints1 = cv.get_polygon_pixels(img, pol1)
        ints2 = cv.get_polygon_pixels(img, pol2)
        if ints1.size < min_samples or ints2.size < min_samples:
            # return a strong penalty score value if there are too few samples
            return 2.0
        _, _, norm_counts1 = _histogram(ints1, normalize=True, bin_size=bin_size)
        bin_centers, bin_edges, norm_counts2 = _histogram(ints2, normalize=True, bin_size=bin_size)
        return _histogram_union(bin_edges, norm_counts1, norm_counts2)

    def plot(
        self,
        h: Optional[float] = None,
        length: float = 2.0,
        offset: float = 0.0,
        camera: bool = False,
        ax: Optional[mpl.axes.Axes] = None,
        cs_kwargs: Dict = None,
        planar=True,
        bottom=True,
        wetted=True,
        swap_y_coords=False,
        planar_kwargs: Dict = PLANAR_KWARGS,
        bottom_kwargs: Dict = BOTTOM_KWARGS,
        wetted_kwargs: Dict = WETTED_KWARGS,
    ) -> mpl.axes.Axes:
        """Plot the cross-section situation.

        This will create a plot of the cross section, with some up to downstream length in a 3D axes. It can
        be combined with a plot of the camera configuration, with `cross_section.camera_config.plot(mode="3d")`.

        Parameters
        ----------
        mode : Literal["camera", "3d"]
            manner in which plotting should be done.
        h : float, optional
            water level [m]. If not provided, the water level is taken from the camera config
            `cross_section.camera_config.gcps["h_ref"]`.
        length : float, optional
            length of the waterline [m], by default 2.0
        offset : float, optional
            perpendicular offset of the waterline from the cross-section [m], by default 0.0
        camera : bool, optional
            If set, return 2D projected polygon, by default False.
        ax : plt.axes, optional
            if not provided, axes is setup (Default: None). If provided, user must take care to provide the correct
            projection. If `camera=False`, an axes must be provided with `projection="3d"`.
        cs_kwargs : dict, optional
            keyword arguments used to make the line plot of the cross section.
        planar : bool, optional
            whether to plot the planar, by default True.
        bottom : bool, optional
            whether to plot the bottom surface, by default True.
        wetted : bool, optional
            whether to plot the wetted surface, by default True.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)
        planar_kwargs : dict, optional
            keyword arguments used to make the polygon plot of the water plane. If not provided, a set of defaults
            will be used that give a natural look.
        bottom_kwargs : dict, optional
            keyword arguments used to make the polygon plot of the bottom surface. If not provided, a set of defaults
            will be used that give a natural look.
        wetted_kwargs : dict, optional
            keyword arguments used to make the polygon plot of the wetted surface. If not provided, a set of defaults
            will be used that give a natural look.

        Returns
        -------
        mpl.axes.Axes
            The developed axes object with all data

        """
        if not cs_kwargs:
            cs_kwargs = {}
        if not ax:
            if not camera:
                ax = plt.axes(projection="3d")
            else:
                ax = plt.axes()
        if h is None:
            h = self.camera_config.gcps["h_ref"]
            warnings.warn(
                "No water level is provided. Camera configuration reference water level is used.", stacklevel=2
            )
        if planar:
            self.plot_planar_surface(
                h=h, length=length, offset=offset, camera=camera, ax=ax, swap_y_coords=swap_y_coords, **planar_kwargs
            )
        if bottom:
            self.plot_bottom_surface(
                length=length, offset=offset, camera=camera, ax=ax, swap_y_coords=swap_y_coords, **bottom_kwargs
            )
        if wetted:
            self.plot_wetted_surface(h=h, camera=camera, ax=ax, swap_y_coords=swap_y_coords, **wetted_kwargs)
        self.plot_cs(ax=ax, camera=camera, swap_y_coords=swap_y_coords, **cs_kwargs)
        # ax.set_aspect("equal")
        return ax

    def plot_cs(self, ax=None, camera=False, swap_y_coords: bool = False, **kwargs):
        """Plot the cross section.

        Parameters
        ----------
        ax : plt.axes, optional
            if not provided, axes is setup (Default: None). If provided, user must take care to provide the correct
            projection. If `camera=False`, an axes must be provided with `projection="3d"`.
        camera : bool, optional
            If set, return 2D projected polygon, by default False.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)
        **kwargs : dict, optional
            keyword arguments used to make the line plot of the cross-section.

        Returns
        -------
        mpl mappable

        """
        if not ax:
            if camera:
                ax = plt.axes()
            else:
                ax = plt.axes(projection="3d")
        if ax.name == "3d" and not camera:
            # map 3d coordinates to x, y, z
            x, y, z = zip(*[(c[0], c[1], c[2]) for c in self.cs_linestring.coords])
            p = ax.plot(x, y, z, **kwargs)
        else:
            if camera:
                # extract pixel locations
                pix = self.camera_config.project_points(
                    list(map(list, self.cs_linestring.coords)), within_image=True, swap_y_coords=swap_y_coords
                )
                # map to x and y arrays
                x, y = zip(*[(c[0], c[1]) for c in pix if np.isfinite(c[0])])
            else:
                x, y = zip(*[(c[0], c[1]) for c in self.cs_linestring_sz.coords])
            p = ax.plot(x, y, **kwargs)
        return p

    def plot_planar_surface(
        self,
        h: float,
        length: float = 2.0,
        offset: float = 0.0,
        camera: bool = False,
        swap_y_coords=False,
        ax=None,
        **kwargs,
    ) -> mpl.axes.Axes:
        """Plot the planar surface for a given water level.

        Parameters
        ----------
        h : float, optional
            water level [m]. If not provided, the water level is taken from the camera config
            `cross_section.camera_config.gcps["h_ref"]`.
        length : float, optional
            length of the waterline [m], by default 2.0
        offset : float, optional
            perpendicular offset of the waterline from the cross-section [m], by default 0.0
        camera : bool, optional
            If set, return 2D projected polygon, by default False.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)
        ax : plt.axes, optional
            if not provided, axes is setup (Default: None). If provided, user must take care to provide the correct
            projection. If `camera=False`, an axes must be provided with `projection="3d"`.
        **kwargs : dict, optional
            keyword arguments used to make the polygon plot of the planar surface.

        Returns
        -------
        plt.axes

        """
        try:
            surf = self.get_planar_surface(
                h=h, length=length, offset=offset, swap_y_coords=swap_y_coords, camera=camera
            )
            if camera:
                p = plot_helpers.plot_polygon(surf, ax=ax, label="surface", **kwargs)
            else:
                p = plot_helpers.plot_3d_polygon(surf, ax=ax, label="surface", **kwargs)
            return p.axes
        except Exception:
            warnings.warn(
                "Cannot plot planar surface as there are too many crossings",
                stacklevel=2,
            )

    def plot_bottom_surface(
        self, length: float = 2.0, offset: float = 0.0, camera: bool = False, ax=None, swap_y_coords=False, **kwargs
    ) -> mpl.axes.Axes:
        """Plot the bottom surface for a given water level.

        Parameters
        ----------
        length : float, optional
            length of the waterline [m], by default 2.0
        offset : float, optional
            perpendicular offset of the waterline from the cross-section [m], by default 0.0
        camera : bool, optional
            If set, return 2D projected polygon, by default False.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)
        ax : plt.axes, optional
            if not provided, axes is setup (Default: None). If provided, user must take care to provide the correct
            projection. If `camera=False`, an axes must be provided with `projection="3d"`.
        **kwargs : dict, optional
            keyword arguments used to make the polygon plot of the bottom surface.

        Returns
        -------
        plt.axes

        """
        surf = self.get_bottom_surface(length=length, offset=offset, camera=camera, swap_y_coords=swap_y_coords)
        if camera:
            p = plot_helpers.plot_polygon(surf, ax=ax, label="bottom", **kwargs)
        else:
            p = plot_helpers.plot_3d_polygon(surf, ax=ax, label="bottom", **kwargs)
        return p.axes

    def plot_wetted_surface(self, h: float, camera: bool = False, swap_y_coords=False, ax=None, **kwargs):
        """Plot the wetted surface for a given water level.

        Parameters
        ----------
        h : float, optional
            water level [m]. If not provided, the water level is taken from the camera config
            `cross_section.camera_config.gcps["h_ref"]`.
        camera : bool, optional
            If set, return 2D projected polygon, by default False.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)
        ax : plt.axes, optional
            if not provided, axes is setup (Default: None). If provided, user must take care to provide the correct
            projection. If `camera=False`, an axes must be provided with `projection="3d"`.
        **kwargs : dict, optional
            keyword arguments used to make the polygon plot of the wetted surface.

        Returns
        -------
        plt.axes

        """
        surf = self.get_wetted_surface(h=h, camera=camera, swap_y_coords=swap_y_coords)
        if camera:
            p = plot_helpers.plot_polygon(surf, ax=ax, label="wetted", **kwargs)
        else:
            p = plot_helpers.plot_3d_polygon(surf, ax=ax, label="wetted", **kwargs)
        return p.axes

    def plot_water_level(
        self,
        h: float,
        length: float = 0.5,
        offset: float = 0.0,
        camera: bool = False,
        swap_y_coords: bool = False,
        add_text: bool = True,
        ax: Optional[mpl.axes.Axes] = None,
        **kwargs,
    ) -> mpl.axes.Axes:
        """Plot the water level at user provided value `h`.

        Parameters
        ----------
        h : float, optional
            water level [m]. If not provided, the water level is taken from the camera config
            `cross_section.camera_config.gcps["h_ref"]`.
        length : float, optional
            length of the waterline [m], by default 2.0
        offset : float, optional
            perpendicular offset of the waterline from the cross-section [m], by default 0.0
        camera : bool, optional
            If set, return 2D projected polygon, by default False.
        swap_y_coords : bool, optional
            if set, all y-coordinates are swapped so that they fit on a flipped version of the image. This is useful
            in case you plot on ascending y-coordinate axis background images (default: False)
        add_text : bool, optional
            Default True, determines whether to add a text label to the water level plot at the farthest line.
        ax : plt.axes, optional
            if not provided, axes is setup (Default: None). If provided, user must take care to provide the correct
            projection. If `camera=False`, an axes must be provided with `projection="3d"`.
        **kwargs : dict, optional
            keyword arguments used to make the line plot of the water level, perpendicular to the cross-section.

        Returns
        -------
        plt.axes

        """
        if ax is None:
            if camera:
                ax = plt.axes()
            else:
                ax = plt.axes(projection="3d")
        lines = self.get_csl_line(h=h, length=length, offset=offset, camera=camera, swap_y_coords=swap_y_coords)
        if camera:
            plot_f = plot_helpers.plot_line
        else:
            plot_f = plot_helpers.plot_3d_line
        for l in lines:
            _ = plot_f(l, ax=ax, zorder=1, **kwargs)
        if add_text and camera:
            points = self.get_csl_point(h=h, camera=False)  # find real-world points
            lens_position_xy = self.camera_config.estimate_lens_position()[0:2]
            dists = [((p.x - lens_position_xy[0]) ** 2 + (p.y - lens_position_xy[1]) ** 2) ** 0.5 for p in points]
            points = self.get_csl_point(h=h, camera=True, swap_y_coords=swap_y_coords)  # find camera positions
            x, y = points[np.argmax(dists)].xy
            x, y = float(x[0]), float(y[0])

            # only plot text in 2D camera perspective at farthest point
            ax.text(
                x, y, "{:1.3f} m.".format(h), path_effects=PATH_EFFECTS, ha="center", va="bottom", size=12, zorder=2
            )
        return ax

    def detect_water_level(
        self,
        img: np.ndarray,
        bank: BANK_OPTIONS = "far",
        bin_size: int = 5,
        length: float = 2.0,
        padding: float = 0.5,
        offset: float = 0.0,
        min_h: Optional[float] = None,
        max_h: Optional[float] = None,
        min_z: Optional[float] = None,
        max_z: Optional[float] = None,
    ) -> float:
        """Detect water level optically from provided image.

        Water level detection is done by first detecting the water line along the cross-section by comparisons
        of distribution functions left and right of hypothesized water lines, and then looking up the water level
        associated with the water line location.

        Parameters
        ----------
        img : np.ndarray
            image (uint8) used to estimate water level from.
        bank: Literal["far", "near", "both"], optional
            select from which bank to detect the water level. Use this if camera is positioned in a way that only
            one shore is clearly distinguishable and not obscured. Typically you will use "far" if the camera is
            positioned on one bank, aimed perpendicular to the flow. Use "near" if not the full cross section is
            visible, but only the part nearest the camera. And leave empty when both banks are clearly visible and
            approximately the same in distance (e.g. middle of a bridge). If not provided, the bank is detected based
            on the best estimate from both banks.
        bin_size : int, optional
            Size of bins for histogram calculation of the provided image intensities, default 5.
        length : float, optional
            length of the waterline [m], by default 2.0
        padding : float, optional
            amount of distance [m] to extend the polygon beyond the waterline, by default 0.5. Two polygons are drawn
            left and right of hypothesized water line at -padding and +padding.
        offset : float, optional
            perpendicular offset of the waterline from the cross-section [m], by default 0.0
        min_h : float, optional
            minimum water level to try detection [m]. If not provided, the minimum water level is taken from the
            cross section.
        max_h : float, optional
            maximum water level to try detection [m]. If not provided, the maximum water level is taken from the
            cross section.
        min_z : float, optional
            same as min_h but using z-coordinates instead of local datum, min_z overrules min_h
        max_z : float, optional
            same as max_z but using z-coordinates instead of local datum, max_z overrules max_h

        """
        if min_z is None:
            if min_h is not None:
                min_z = self.camera_config.h_to_z(min_h)
                min_z = np.maximum(min_z, self.z.min())
        if max_z is None:
            if max_h is not None:
                max_z = self.camera_config.h_to_z(max_h)
                max_z = np.minimum(max_z, self.z.max())
        if min_z and max_z:
            if min_z > max_z:
                raise ValueError("Minimum water level is higher than maximum water level.")

        if len(img.shape) == 3:
            # flatten image first if it his a time dimension
            img = img.mean(axis=2)
        assert (
            img.shape[0] == self.camera_config.height
        ), f"Image height {img.shape[0]} is not the same as camera_config height {self.camera_config.height}"
        assert (
            img.shape[1] == self.camera_config.width
        ), f"Image width {img.shape[1]} is not the same as camera_config width {self.camera_config.width}"
        # determine the relevant start point if only one is used
        # import pdb;pdb.set_trace()
        l_min, l_max = self.get_line_of_interest(bank=bank)
        opt = differential_evolution(
            self.get_histogram_score,
            popsize=50,
            bounds=[(l_min, l_max)],
            args=(img, bin_size, offset, padding, length, min_z, max_z),
            atol=0.01,  # one mm
        )
        z = self.interp_z(opt.x[0])
        h = self.camera_config.z_to_h(z)
        # warning if the optimum is on the edge of the search space for l
        if np.isclose(opt.x[0], l_min) or np.isclose(opt.x[0], l_max):
            warnings.warn(
                f"The detected water level is on the edge of the search space and may be wrong. "
                f"Water level is {h} m. at cross-section length {opt.x[0]}.",
                UserWarning,
                stacklevel=2,
            )
        return h



"""PIV processing wrappers for OpenPIV."""

from typing import List, Optional, Tuple, Union

import numpy as np
import openpiv.pyprocess
import openpiv.tools
import xarray as xr

__all__ = [
    "get_openpiv",
    "piv",
]


def get_openpiv(frames, y, x, dt, **kwargs):
    """Compute time-resolved Particle Image Velocimetry (PIV) using Fast Fourier Transform (FFT) within OpenPIV.

    Calculates velocity using the OpenPIV algorithms by processing sequential frames
    from a dataset and returning the velocity components, signal-to-noise ratio, and
    correlation values. The function shifts frames in time and applies the PIV algorithm
    to compute flow fields over the specified spatial axes.

    Parameters
    ----------
    frames : xarray.Dataset
        The input dataset containing time-dependent frames with coordinates.
    y : array-like
        The spatial coordinates along the y-axis where the outputs should be interpolated.
    x : array-like
        The spatial coordinates along the x-axis where the outputs should be interpolated.
    dt : float
        The time step between consecutive frames (used to go from per-frame to per-second displacement).
    **kwargs : dict
        Additional keyword arguments to be passed to the PIV function.

    Returns
    -------
    xarray.Dataset
        A dataset containing computed velocity components `v_x` and `v_y`,
        signal-to-noise ratios `s2n`, and correlation values `corr`. The dataset
        includes updated x and y coordinates representing the flow field grid.

    """
    # first get rid of coordinates that need to be recalculated
    coords_drop = list(set(frames.coords) - set(frames.dims))
    frames = frames.drop_vars(coords_drop)
    # get frames and shifted frames in time
    frames1 = frames.shift(time=1)[1:].chunk({"time": 1})
    frames2 = frames[1:].chunk({"time": 1})
    # retrieve all data arrays
    v_x, v_y, s2n, corr = xr.apply_ufunc(
        piv,
        frames1,
        frames2,
        dt,
        kwargs=kwargs,
        input_core_dims=[["y", "x"], ["y", "x"], []],
        output_core_dims=[["new_y", "new_x"]] * 4,
        dask_gufunc_kwargs={
            "output_sizes": {"new_y": len(y), "new_x": len(x)},
        },
        output_dtypes=[np.float32] * 4,
        vectorize=True,
        keep_attrs=True,
        dask="parallelized",
    )
    # merge all DataArrays in one Dataset
    ds = xr.merge([v_x.rename("v_x"), v_y.rename("v_y"), s2n.rename("s2n"), corr.rename("corr")]).rename(
        {"new_x": "x", "new_y": "y"}
    )
    # add y and x-axis values
    ds["y"] = y
    ds["x"] = x
    return ds


def piv(
    frame_a,
    frame_b,
    dt,
    res_x=0.01,
    res_y=0.01,
    search_area_size=30,
    window_size=None,
    overlap=None,
    **kwargs,
):
    """Perform PIV analysis on two sequential frames following keyword arguments from openpiv.

    This function also computes the correlations per interrogation window, so that poorly correlated values can be
    filtered out. Furthermore, the resolution is used to convert pixel per second velocity estimates, into meter per
    second velocity estimates. The centre of search area columns and rows are also returned so that a georeferenced
    grid can be written from the results.

    Note: Typical openpiv kwargs are for instance
    window_size=60, overlap=30, search_area_size=60, dt=1./25

    Parameters
    ----------
    frame_a: np.ndarray (2D)
        first frame
    frame_b: np.ndarray (2D)
        second frame
    dt : float
        time resolution in seconds.
    res_x: float, optional
        resolution of x-dir pixels in a user-defined unit per pixel (e.g. m pixel-1) Default: 0.01
    res_y: float, optional
        resolution of y-dir pixels in a user-defined unit per pixel (e.g. m pixel-1) Default: 0.01
    search_area_size: int, optional
        length of subsetted matrix to search for correlations (default: 30)
    window_size: int, optional
        size of interrogation window in amount of pixels. If not set, it is set equal to search_area_size
        (default: None).
    overlap: int, optional
        length of overlap between interrogation windows. If not set, this defaults to 50% of the window_size parameter
        (default: None).
    **kwargs: dict
        keyword arguments related to openpiv. See openpiv manual for further information

    Returns
    -------
    v_x: np.ndarray(2D)
        raw x-dir velocities [m s-1] in interrogation windows (requires filtering to get valid velocities)
    v_y: np.ndarray (2D)
        raw y-dir velocities [m s-1] in interrogation windows (requires filtering to get valid velocities)
    s2n: np.ndarray (2D)
        signal to noise ratio, measured as maximum correlation found divided by the mean correlation
        (method="peak2mean") or second to maximum correlation (method="peak2peak") found within search area
    corr: np.ndarray (2D)
        correlation values in interrogation windows

    """
    window_size = search_area_size if window_size is None else window_size
    overlap = int(round(window_size) / 2) if overlap is None else overlap
    # modified version of extended_search_area_piv to accomodate exporting corr
    v_x, v_y, s2n, corr = extended_search_area_piv(
        frame_a, frame_b, dt=dt, search_area_size=search_area_size, overlap=overlap, window_size=window_size, **kwargs
    )
    return v_x * res_x, v_y * res_y, s2n, corr


def extended_search_area_piv(
    frame_a: np.ndarray,
    frame_b: np.ndarray,
    window_size: int,
    overlap: int = 0,
    dt: float = 1.0,
    search_area_size: Optional[Union[Tuple[int, int], List[int], int]] = None,
    correlation_method: str = "circular",
    subpixel_method: str = "gaussian",
    sig2noise_method: Optional[str] = "peak2mean",
    width: int = 2,
    normalized_correlation: bool = True,
    use_vectorized: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Perform PIV cross-correlation analysis.

    Extended area search can be used to increased dynamic range. The search region
    in the second frame is larger than the interrogation window size in the
    first frame. For Cython implementation see
    openpiv.process.extended_search_area_piv

    This is a pure python implementation of the standard PIV cross-correlation
    algorithm. It is a zero order displacement predictor, and no iterative
    process is performed.

    Parameters
    ----------
    frame_a : 2d np.ndarray
        an two dimensions array of integers containing grey levels of
        the first frame.

    frame_b : 2d np.ndarray
        an two dimensions array of integers containing grey levels of
        the second frame.

    window_size : int
        the size of the (square) interrogation window, [default: 32 pix].

    overlap : int
        the number of pixels by which two adjacent windows overlap
        [default: 16 pix].

    dt : float
        the time delay separating the two frames [default: 1.0].

    correlation_method : string
        one of the two methods implemented: 'circular' or 'linear',
        default: 'circular', it's faster, without zero-padding
        'linear' requires also normalized_correlation = True (see below)

    subpixel_method : string
         one of the following methods to estimate subpixel location of the
         peak:
         'centroid' [replaces default if correlation map is negative],
         'gaussian' [default if correlation map is positive],
         'parabolic'.

    sig2noise_method : string
        defines the method of signal-to-noise-ratio measure,
        ('peak2peak' or 'peak2mean'. If None, no measure is performed.)

    width : int
        the half size of the region around the first
        correlation peak to ignore for finding the second
        peak. [default: 2]. Only used if ``sig2noise_method==peak2peak``.

    search_area_size : int
       the size of the interrogation window in the second frame,
       default is the same interrogation window size and it is a
       fallback to the simplest FFT based PIV

    normalized_correlation: bool
        if True, then the image intensity will be modified by removing
        the mean, dividing by the standard deviation and
        the correlation map will be normalized. It's slower but could be
        more robust

    use_vectorized : bool
        If set, vectorization is used to speed up analysis.

    Returns
    -------
    u : 2d np.ndarray
        a two dimensional array containing the u velocity component,
        in pixels/seconds.

    v : 2d np.ndarray
        a two dimensional array containing the v velocity component,
        in pixels/seconds.

    sig2noise : 2d np.ndarray ( optional: only if sig2noise_method != None )
        a two dimensional array the signal to noise ratio for each
        window pair.

    corr : 2d np.ndarray
        a two dimensional array with the maximum correlation values found in each interrogation window.

    The implementation of the one-step direct correlation with different
    size of the interrogation window and the search area. The increased
    size of the search areas cope with the problem of loss of pairs due
    to in-plane motion, allowing for a smaller interrogation window size,
    without increasing the number of outlier vectors.

    See:

    Particle-Imaging Techniques for Experimental Fluid Mechanics

    Annual Review of Fluid Mechanics
    Vol. 23: 261-304 (Volume publication date January 1991)
    DOI: 10.1146/annurev.fl.23.010191.001401

    originally implemented in process.pyx in Cython and converted to
    a NumPy vectorized solution in pyprocess.py

    """
    if search_area_size is not None:
        if isinstance(search_area_size, tuple) == False and isinstance(search_area_size, list) == False:
            search_area_size = [search_area_size, search_area_size]
    if isinstance(window_size, tuple) == False and isinstance(window_size, list) == False:
        window_size = [window_size, window_size]
    if isinstance(overlap, tuple) == False and isinstance(overlap, list) == False:
        overlap = [overlap, overlap]

    # check the inputs for validity
    search_area_size = window_size if search_area_size is None else search_area_size

    if overlap[0] >= window_size[0] or overlap[1] >= window_size[1]:
        raise ValueError("Overlap has to be smaller than the window_size")

    if search_area_size[0] < window_size[0] or search_area_size[1] < window_size[1]:
        raise ValueError("Search size cannot be smaller than the window_size")

    if (window_size[1] > frame_a.shape[0]) or (window_size[0] > frame_a.shape[1]):
        raise ValueError("window size cannot be larger than the image")

    # get field shape
    n_rows, n_cols = openpiv.pyprocess.get_field_shape(frame_a.shape, search_area_size, overlap)

    # We implement the new vectorized code
    aa = openpiv.pyprocess.sliding_window_array(frame_a, search_area_size, overlap)
    bb = openpiv.pyprocess.sliding_window_array(frame_b, search_area_size, overlap)

    # for the case of extended seearch, the window size is smaller than
    # the search_area_size. In order to keep it all vectorized the
    # approach is to use the interrogation window in both
    # frames of the same size of search_area_asize,
    # but mask out the region around
    # the interrogation window in the frame A

    if search_area_size > window_size:
        # before masking with zeros we need to remove
        # edges

        aa = openpiv.pyprocess.normalize_intensity(aa)
        bb = openpiv.pyprocess.normalize_intensity(bb)

        mask = np.zeros((search_area_size[0], search_area_size[1])).astype(aa.dtype)
        pady = int((search_area_size[0] - window_size[0]) / 2)
        padx = int((search_area_size[1] - window_size[1]) / 2)
        mask[slice(pady, search_area_size[0] - pady), slice(padx, search_area_size[1] - padx)] = 1
        mask = np.broadcast_to(mask, aa.shape)
        aa *= mask

    corr = openpiv.pyprocess.fft_correlate_images(
        aa, bb, correlation_method=correlation_method, normalized_correlation=normalized_correlation
    )
    if use_vectorized == True:
        u, v = openpiv.pyprocess.vectorized_correlation_to_displacements(
            corr, n_rows, n_cols, subpixel_method=subpixel_method
        )
    else:
        u, v = openpiv.pyprocess.correlation_to_displacement(corr, n_rows, n_cols, subpixel_method=subpixel_method)

    # return output depending if user wanted sig2noise information
    sig2noise = np.zeros_like(u) * np.nan
    if sig2noise_method is not None:
        if use_vectorized == True:
            sig2noise = openpiv.pyprocess.vectorized_sig2noise_ratio(
                corr, sig2noise_method=sig2noise_method, width=width
            )
        else:
            sig2noise = openpiv.pyprocess.sig2noise_ratio(corr, sig2noise_method=sig2noise_method, width=width)

    sig2noise = sig2noise.reshape(n_rows, n_cols)
    # extended code for exporting the maximum found value for corr
    corr = corr.max(axis=-1).max(axis=-1).reshape((n_rows, n_cols))

    return u / dt, v / dt, sig2noise, corr



"""pyorc projection functions."""

from __future__ import annotations

from typing import Any, Optional

import cv2
import numba as nb
import numpy as np
import xarray as xr

from . import cv

# from . import CameraConfig

__all__ = ["project_numpy", "project_cv"]


@nb.njit(
    nb.float32[:](nb.float32[:], nb.int64[:], nb.int64),
    nogil=True,
    parallel=False,
    cache=True,
)
def _group_average(data, idx, num_groups):
    """Compute group averages on sampled data, using unique values in idx as group indices.

    Parameters
    ----------
    data : np.ndarray
        1d-array containing values sampled from the original data
    idx : np.ndarray
        1d-array containing indices of the original data that correspond to the values in data.
        idx has the same size as data
    num_groups : int
        the amount of groups to average over. I.e. result of np.unique(idx).size

    """
    # Arrays to hold the sum and count for each group
    group_sums = np.zeros(num_groups, dtype=np.float32)
    group_counts = np.zeros(num_groups, dtype=np.int64)

    # Accumulate sums and counts for each group
    for i in range(len(data)):
        group_sums[idx[i]] += data[i]
        group_counts[idx[i]] += 1

    # Compute the averages
    averages = np.zeros(num_groups, dtype=np.float32)
    for group in range(num_groups):
        # if group_counts[group] > 0:  # Avoid division by zero
        averages[group] = group_sums[group] / group_counts[group]
    return averages


def project_cv(da: xr.DataArray, cc: Any, x: np.ndarray, y: np.ndarray, z: np.ndarray, reducer: str):
    """Projection method that uses pure OpenCV.

    Reprojection is done in two steps: undistortion and reprojection.
    This method gives incorrect mapping in case of very strong distortion and/or where part of the area of interest is
    outside of the field of view. In these cases, it is strongly recommended to use `project_numpy` instead.

    Parameters
    ----------
    da : xr.DataArray
        Frames time series
    cc : pyorc.CameraConfig
        pyorc CameraConfig object
    x : np.ndarray
        x-axis of target ortho image
    y : np.ndarray
        y-axis of target ortho image
    z : float
        vertical level value in real-world coordinates
    reducer : str
        not used, only passed to make function commensurate with project_numpy

    Returns
    -------
    da_proj : xr.DataArray
        Reprojected Frames

    """
    da_undistort = xr.apply_ufunc(
        cv.undistort_img,
        da,
        input_core_dims=[["y", "x"]],
        output_core_dims=[["y", "x"]],
        output_dtypes=da.dtype,
        kwargs={"camera_matrix": cc.camera_matrix, "dist_coeffs": cc.dist_coeffs},
        dask="parallelized",
        keep_attrs=True,
        vectorize=True,
    )  # .rename({
    # also undistort the src control points
    cc.gcps["src"] = cv.undistort_points(cc.gcps["src"], cc.camera_matrix, cc.dist_coeffs)
    h_a = cc.z_to_h(z)
    src = cc.get_bbox(mode="camera", h_a=h_a, expand_exterior=False).exterior.coords[0:4]
    dst_xy = cc.get_bbox(expand_exterior=False).exterior.coords[0:4]
    # get geographic coordinates bbox corners
    dst = cv.transform_to_bbox(dst_xy, cc.bbox, cc.resolution)
    M = cv.get_M_2D(src, dst)
    da_proj = xr.apply_ufunc(
        cv.get_ortho,
        da_undistort,
        kwargs={"M": M, "shape": tuple(np.flipud(cc.shape)), "flags": cv2.INTER_AREA},
        input_core_dims=[["y", "x"]],
        output_core_dims=[["new_y", "new_x"]],
        dask_gufunc_kwargs={
            "output_sizes": {"new_y": len(y), "new_x": len(x)},
        },
        output_dtypes=[da.dtype],
        vectorize=True,
        exclude_dims=set(("y", "x")),
        dask="parallelized",
        keep_attrs=True,
    ).rename({"new_y": "y", "new_x": "x"})
    da_proj["y"] = y
    da_proj["x"] = x
    return da_proj


def img_to_ortho(img, x, y, idx_img, idx_ortho, src_idx=None, uidx=None, norm_idx=None):
    """Project from original image to ortho image using pre-calculated index mapping.

    This function can use nearest neighbour (for undersampled areas)  as well as averages (in oversampled areas).
    If `src_idx` is provided, then the averages are computed using the values in `src_idx` as group indices.

    Parameters
    ----------
    img : np.ndarray
        2d array containing the original image
    x : np.ndarray
        x-axis of target ortho image
    y : np.ndarray
        y-axis of target ortho image
    idx_img : np.ndarray[int]
        indices of the original image that correspond to the values in img. Used for nearest neighbour.
    idx_ortho : np.ndarray[int]
        indixes of the target ortho image that correspond to the values in img. Used for nearest neighbour.
    src_idx : np.ndarray
        1D array of flattened indices of the source image pixels that correspond
        to selected orthographic grid pixels.
    uidx : np.ndarray
        Sorted unique indices of the filtered orthographic grid pixels.
    norm_idx : np.ndarray
        Normalized indices corresponding to their positions in the unique
        filtered orthographic grid pixels. Used for averaging.

    """
    img = np.float32(img.flatten())
    # first make new flattened image
    new_arr = np.zeros((len(y) * len(x)))
    new_arr[idx_ortho] = img[idx_img]
    if src_idx is not None:
        samples_for_mean = img[src_idx]
        # Compute Group Averages
        averages = _group_average(samples_for_mean, norm_idx, len(uidx))
        new_arr[uidx] = averages
    # reshape to original grid
    return new_arr.reshape(len(y), -1)


def project_numpy(
    da: xr.DataArray,
    cc: Any,
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    reducer: Optional[str] = "mean",
):
    """Project from FOV pixels directly to target grid, including undistortion and projection.

    Parameters
    ----------
    da : xr.DataArray
        Frames time series
    cc : pyorc.CameraConfig
        pyorc CameraConfig object
    x : np.ndarray
        x-axis of target ortho image
    y : np.ndarray
        y-axis of target ortho image
    z : float
        vertical level value in real-world coordinates
    reducer : str, optional
        If set to a valid reducer (like mean, median, max) oversampled target pixels will be reduced by using the set
        reducer. Oversampled target pixels are defined as pixels that have more than one pixels in the original
        Field of View that fit within that pixel. All other pixels are defined with nearest-neighbour. Default is
        "nearest" for nearest neighbour.

    Returns
    -------
    da_proj : xr.DataArray
        Reprojected Frames

    """
    # create coordinate system for target grid
    idx_img_nn, idx_ortho_nn = cc.map_idx_img_ortho(x, y, z)
    if reducer == "mean":
        src_idx_mean, uidx_mean, norm_idx_mean = cc.map_mean_idx_img_ortho(x, y, z)
    else:
        src_idx_mean, uidx_mean, norm_idx_mean = None, None, None
    # da.load()
    da_proj = xr.apply_ufunc(
        img_to_ortho,
        da,
        kwargs={
            "x": x,
            "y": y,
            "idx_img": idx_img_nn,
            "idx_ortho": idx_ortho_nn,
            "src_idx": src_idx_mean,
            "uidx": uidx_mean,
            "norm_idx": norm_idx_mean,
        },
        input_core_dims=[["y", "x"]],
        output_core_dims=[["new_y", "new_x"]],
        dask_gufunc_kwargs={
            "output_sizes": {"new_y": len(y), "new_x": len(x)},
        },
        output_dtypes=[da.dtype],
        vectorize=True,
        exclude_dims=set(("y", "x")),
        dask="parallelized",
        keep_attrs=True,
    ).rename({"new_y": "y", "new_x": "x"})
    da_proj["y"] = y
    da_proj["x"] = x
    return da_proj
